words = [
    "abandon",
    "ability",
    "able",
    "about",
    "above",
    "absent",
    "absorb",
    "abstract",
    "absurd",
    "abuse",
    "access",
    "accident",
    "account",
    "accuse",
    "achieve",
    "acid",
    "acoustic",
    "acquire",
    "across",
    "act",
    "action",
    "actor",
    "actress",
    "actual",
    "adapt",
    "add",
    "addict",
    "address",
    "adjust",
    "admit",
    "adult",
    "advance",
    "advice",
    "aerobic",
    "affair",
    "afford",
    "afraid",
    "again",
    "age",
    "agent",
    "agree",
    "ahead",
    "aim",
    "air",
    "airport",
    "aisle",
    "alarm",
    "album",
    "alcohol",
    "alert",
    "alien",
    "all",
    "alley",
    "allow",
    "almost",
    "alone",
    "alpha",
    "already",
    "also",
    "alter",
    "always",
    "amateur",
    "amazing",
    "among",
    "amount",
    "amused",
    "analyst",
    "anchor",
    "ancient",
    "anger",
    "angle",
    "angry",
    "animal",
    "ankle",
    "announce",
    "annual",
    "another",
    "answer",
    "antenna",
    "antique",
    "anxiety",
    "any",
    "apart",
    "apology",
    "appear",
    "apple",
    "approve",
    "april",
    "arch",
    "arctic",
    "area",
    "arena",
    "argue",
    "arm",
    "armed",
    "armor",
    "army",
    "around",
    "arrange",
    "arrest",
    "arrive",
    "arrow",
    "art",
    "artefact",
    "artist",
    "artwork",
    "ask",
    "aspect",
    "assault",
    "asset",
    "assist",
    "assume",
    "asthma",
    "athlete",
    "atom",
    "attack",
    "attend",
    "attitude",
    "attract",
    "auction",
    "audit",
    "august",
    "aunt",
    "author",
    "auto",
    "autumn",
    "average",
    "avocado",
    "avoid",
    "awake",
    "aware",
    "away",
    "awesome",
    "awful",
    "awkward",
    "axis",
    "baby",
    "bachelor",
    "bacon",
    "badge",
    "bag",
    "balance",
    "balcony",
    "ball",
    "bamboo",
    "banana",
    "banner",
    "bar",
    "barely",
    "bargain",
    "barrel",
    "base",
    "basic",
    "basket",
    "battle",
    "beach",
    "bean",
    "beauty",
    "because",
    "become",
    "beef",
    "before",
    "begin",
    "behave",
    "behind",
    "believe",
    "below",
    "belt",
    "bench",
    "benefit",
    "best",
    "betray",
    "better",
    "between",
    "beyond",
    "bicycle",
    "bid",
    "bike",
    "bind",
    "biology",
    "bird",
    "birth",
    "bitter",
    "black",
    "blade",
    "blame",
    "blanket",
    "blast",
    "bleak",
    "bless",
    "blind",
    "blood",
    "blossom",
    "blouse",
    "blue",
    "blur",
    "blush",
    "board",
    "boat",
    "body",
    "boil",
    "bomb",
    "bone",
    "bonus",
    "book",
    "boost",
    "border",
    "boring",
    "borrow",
    "boss",
    "bottom",
    "bounce",
    "box",
    "boy",
    "bracket",
    "brain",
    "brand",
    "brass",
    "brave",
    "bread",
    "breeze",
    "brick",
    "bridge",
    "brief",
    "bright",
    "bring",
    "brisk",
    "broccoli",
    "broken",
    "bronze",
    "broom",
    "brother",
    "brown",
    "brush",
    "bubble",
    "buddy",
    "budget",
    "buffalo",
    "build",
    "bulb",
    "bulk",
    "bullet",
    "bundle",
    "bunker",
    "burden",
    "burger",
    "burst",
    "bus",
    "business",
    "busy",
    "butter",
    "buyer",
    "buzz",
    "cabbage",
    "cabin",
    "cable",
    "cactus",
    "cage",
    "cake",
    "call",
    "calm",
    "camera",
    "camp",
    "can",
    "canal",
    "cancel",
    "candy",
    "cannon",
    "canoe",
    "canvas",
    "canyon",
    "capable",
    "capital",
    "captain",
    "car",
    "carbon",
    "card",
    "cargo",
    "carpet",
    "carry",
    "cart",
    "case",
    "cash",
    "casino",
    "castle",
    "casual",
    "cat",
    "catalog",
    "catch",
    "category",
    "cattle",
    "caught",
    "cause",
    "caution",
    "cave",
    "ceiling",
    "celery",
    "cement",
    "census",
    "century",
    "cereal",
    "certain",
    "chair",
    "chalk",
    "champion",
    "change",
    "chaos",
    "chapter",
    "charge",
    "chase",
    "chat",
    "cheap",
    "check",
    "cheese",
    "chef",
    "cherry",
    "chest",
    "chicken",
    "chief",
    "child",
    "chimney",
    "choice",
    "choose",
    "chronic",
    "chuckle",
    "chunk",
    "churn",
    "cigar",
    "cinnamon",
    "circle",
    "citizen",
    "city",
    "civil",
    "claim",
    "clap",
    "clarify",
    "claw",
    "clay",
    "clean",
    "clerk",
    "clever",
    "click",
    "client",
    "cliff",
    "climb",
    "clinic",
    "clip",
    "clock",
    "clog",
    "close",
    "cloth",
    "cloud",
    "clown",
    "club",
    "clump",
    "cluster",
    "clutch",
    "coach",
    "coast",
    "coconut",
    "code",
    "coffee",
    "coil",
    "coin",
    "collect",
    "color",
    "column",
    "combine",
    "come",
    "comfort",
    "comic",
    "common",
    "company",
    "concert",
    "conduct",
    "confirm",
    "congress",
    "connect",
    "consider",
    "control",
    "convince",
    "cook",
    "cool",
    "copper",
    "copy",
    "coral",
    "core",
    "corn",
    "correct",
    "cost",
    "cotton",
    "couch",
    "country",
    "couple",
    "course",
    "cousin",
    "cover",
    "coyote",
    "crack",
    "cradle",
    "craft",
    "cram",
    "crane",
    "crash",
    "crater",
    "crawl",
    "crazy",
    "cream",
    "credit",
    "creek",
    "crew",
    "cricket",
    "crime",
    "crisp",
    "critic",
    "crop",
    "cross",
    "crouch",
    "crowd",
    "crucial",
    "cruel",
    "cruise",
    "crumble",
    "crunch",
    "crush",
    "cry",
    "crystal",
    "cube",
    "culture",
    "cup",
    "cupboard",
    "curious",
    "current",
    "curtain",
    "curve",
    "cushion",
    "custom",
    "cute",
    "cycle",
    "dad",
    "damage",
    "damp",
    "dance",
    "danger",
    "daring",
    "dash",
    "daughter",
    "dawn",
    "day",
    "deal",
    "debate",
    "debris",
    "decade",
    "december",
    "decide",
    "decline",
    "decorate",
    "decrease",
    "deer",
    "defense",
    "define",
    "defy",
    "degree",
    "delay",
    "deliver",
    "demand",
    "demise",
    "denial",
    "dentist",
    "deny",
    "depart",
    "depend",
    "deposit",
    "depth",
    "deputy",
    "derive",
    "describe",
    "desert",
    "design",
    "desk",
    "despair",
    "destroy",
    "detail",
    "detect",
    "develop",
    "device",
    "devote",
    "diagram",
    "dial",
    "diamond",
    "diary",
    "dice",
    "diesel",
    "diet",
    "differ",
    "digital",
    "dignity",
    "dilemma",
    "dinner",
    "dinosaur",
    "direct",
    "dirt",
    "disagree",
    "discover",
    "disease",
    "dish",
    "dismiss",
    "disorder",
    "display",
    "distance",
    "divert",
    "divide",
    "divorce",
    "dizzy",
    "doctor",
    "document",
    "dog",
    "doll",
    "dolphin",
    "domain",
    "donate",
    "donkey",
    "donor",
    "door",
    "dose",
    "double",
    "dove",
    "draft",
    "dragon",
    "drama",
    "drastic",
    "draw",
    "dream",
    "dress",
    "drift",
    "drill",
    "drink",
    "drip",
    "drive",
    "drop",
    "drum",
    "dry",
    "duck",
    "dumb",
    "dune",
    "during",
    "dust",
    "dutch",
    "duty",
    "dwarf",
    "dynamic",
    "eager",
    "eagle",
    "early",
    "earn",
    "earth",
    "easily",
    "east",
    "easy",
    "echo",
    "ecology",
    "economy",
    "edge",
    "edit",
    "educate",
    "effort",
    "egg",
    "eight",
    "either",
    "elbow",
    "elder",
    "electric",
    "elegant",
    "element",
    "elephant",
    "elevator",
    "elite",
    "else",
    "embark",
    "embody",
    "embrace",
    "emerge",
    "emotion",
    "employ",
    "empower",
    "empty",
    "enable",
    "enact",
    "end",
    "endless",
    "endorse",
    "enemy",
    "energy",
    "enforce",
    "engage",
    "engine",
    "enhance",
    "enjoy",
    "enlist",
    "enough",
    "enrich",
    "enroll",
    "ensure",
    "enter",
    "entire",
    "entry",
    "envelope",
    "episode",
    "equal",
    "equip",
    "era",
    "erase",
    "erode",
    "erosion",
    "error",
    "erupt",
    "escape",
    "essay",
    "essence",
    "estate",
    "eternal",
    "ethics",
    "evidence",
    "evil",
    "evoke",
    "evolve",
    "exact",
    "example",
    "excess",
    "exchange",
    "excite",
    "exclude",
    "excuse",
    "execute",
    "exercise",
    "exhaust",
    "exhibit",
    "exile",
    "exist",
    "exit",
    "exotic",
    "expand",
    "expect",
    "expire",
    "explain",
    "expose",
    "express",
    "extend",
    "extra",
    "eye",
    "eyebrow",
    "fabric",
    "face",
    "faculty",
    "fade",
    "faint",
    "faith",
    "fall",
    "false",
    "fame",
    "family",
    "famous",
    "fan",
    "fancy",
    "fantasy",
    "farm",
    "fashion",
    "fat",
    "fatal",
    "father",
    "fatigue",
    "fault",
    "favorite",
    "feature",
    "february",
    "federal",
    "fee",
    "feed",
    "feel",
    "female",
    "fence",
    "festival",
    "fetch",
    "fever",
    "few",
    "fiber",
    "fiction",
    "field",
    "figure",
    "file",
    "film",
    "filter",
    "final",
    "find",
    "fine",
    "finger",
    "finish",
    "fire",
    "firm",
    "first",
    "fiscal",
    "fish",
    "fit",
    "fitness",
    "fix",
    "flag",
    "flame",
    "flash",
    "flat",
    "flavor",
    "flee",
    "flight",
    "flip",
    "float",
    "flock",
    "floor",
    "flower",
    "fluid",
    "flush",
    "fly",
    "foam",
    "focus",
    "fog",
    "foil",
    "fold",
    "follow",
    "food",
    "foot",
    "force",
    "forest",
    "forget",
    "fork",
    "fortune",
    "forum",
    "forward",
    "fossil",
    "foster",
    "found",
    "fox",
    "fragile",
    "frame",
    "frequent",
    "fresh",
    "friend",
    "fringe",
    "frog",
    "front",
    "frost",
    "frown",
    "frozen",
    "fruit",
    "fuel",
    "fun",
    "funny",
    "furnace",
    "fury",
    "future",
    "gadget",
    "gain",
    "galaxy",
    "gallery",
    "game",
    "gap",
    "garage",
    "garbage",
    "garden",
    "garlic",
    "garment",
    "gas",
    "gasp",
    "gate",
    "gather",
    "gauge",
    "gaze",
    "general",
    "genius",
    "genre",
    "gentle",
    "genuine",
    "gesture",
    "ghost",
    "giant",
    "gift",
    "giggle",
    "ginger",
    "giraffe",
    "girl",
    "give",
    "glad",
    "glance",
    "glare",
    "glass",
    "glide",
    "glimpse",
    "globe",
    "gloom",
    "glory",
    "glove",
    "glow",
    "glue",
    "goat",
    "goddess",
    "gold",
    "good",
    "goose",
    "gorilla",
    "gospel",
    "gossip",
    "govern",
    "gown",
    "grab",
    "grace",
    "grain",
    "grant",
    "grape",
    "grass",
    "gravity",
    "great",
    "green",
    "grid",
    "grief",
    "grit",
    "grocery",
    "group",
    "grow",
    "grunt",
    "guard",
    "guess",
    "guide",
    "guilt",
    "guitar",
    "gun",
    "gym",
    "habit",
    "hair",
    "half",
    "hammer",
    "hamster",
    "hand",
    "happy",
    "harbor",
    "hard",
    "harsh",
    "harvest",
    "hat",
    "have",
    "hawk",
    "hazard",
    "head",
    "health",
    "heart",
    "heavy",
    "hedgehog",
    "height",
    "hello",
    "helmet",
    "help",
    "hen",
    "hero",
    "hidden",
    "high",
    "hill",
    "hint",
    "hip",
    "hire",
    "history",
    "hobby",
    "hockey",
    "hold",
    "hole",
    "holiday",
    "hollow",
    "home",
    "honey",
    "hood",
    "hope",
    "horn",
    "horror",
    "horse",
    "hospital",
    "host",
    "hotel",
    "hour",
    "hover",
    "hub",
    "huge",
    "human",
    "humble",
    "humor",
    "hundred",
    "hungry",
    "hunt",
    "hurdle",
    "hurry",
    "hurt",
    "husband",
    "hybrid",
    "ice",
    "icon",
    "idea",
    "identify",
    "idle",
    "ignore",
    "ill",
    "illegal",
    "illness",
    "image",
    "imitate",
    "immense",
    "immune",
    "impact",
    "impose",
    "improve",
    "impulse",
    "inch",
    "include",
    "income",
    "increase",
    "index",
    "indicate",
    "indoor",
    "industry",
    "infant",
    "inflict",
    "inform",
    "inhale",
    "inherit",
    "initial",
    "inject",
    "injury",
    "inmate",
    "inner",
    "innocent",
    "input",
    "inquiry",
    "insane",
    "insect",
    "inside",
    "inspire",
    "install",
    "intact",
    "interest",
    "into",
    "invest",
    "invite",
    "involve",
    "iron",
    "island",
    "isolate",
    "issue",
    "item",
    "ivory",
    "jacket",
    "jaguar",
    "jar",
    "jazz",
    "jealous",
    "jeans",
    "jelly",
    "jewel",
    "job",
    "join",
    "joke",
    "journey",
    "joy",
    "judge",
    "juice",
    "jump",
    "jungle",
    "junior",
    "junk",
    "just",
    "kangaroo",
    "keen",
    "keep",
    "ketchup",
    "key",
    "kick",
    "kid",
    "kidney",
    "kind",
    "kingdom",
    "kiss",
    "kit",
    "kitchen",
    "kite",
    "kitten",
    "kiwi",
    "knee",
    "knife",
    "knock",
    "know",
    "lab",
    "label",
    "labor",
    "ladder",
    "lady",
    "lake",
    "lamp",
    "language",
    "laptop",
    "large",
    "later",
    "latin",
    "laugh",
    "laundry",
    "lava",
    "law",
    "lawn",
    "lawsuit",
    "layer",
    "lazy",
    "leader",
    "leaf",
    "learn",
    "leave",
    "lecture",
    "left",
    "leg",
    "legal",
    "legend",
    "leisure",
    "lemon",
    "lend",
    "length",
    "lens",
    "leopard",
    "lesson",
    "letter",
    "level",
    "liar",
    "liberty",
    "library",
    "license",
    "life",
    "lift",
    "light",
    "like",
    "limb",
    "limit",
    "link",
    "lion",
    "liquid",
    "list",
    "little",
    "live",
    "lizard",
    "load",
    "loan",
    "lobster",
    "local",
    "lock",
    "logic",
    "lonely",
    "long",
    "loop",
    "lottery",
    "loud",
    "lounge",
    "love",
    "loyal",
    "lucky",
    "luggage",
    "lumber",
    "lunar",
    "lunch",
    "luxury",
    "lyrics",
    "machine",
    "mad",
    "magic",
    "magnet",
    "maid",
    "mail",
    "main",
    "major",
    "make",
    "mammal",
    "man",
    "manage",
    "mandate",
    "mango",
    "mansion",
    "manual",
    "maple",
    "marble",
    "march",
    "margin",
    "marine",
    "market",
    "marriage",
    "mask",
    "mass",
    "master",
    "match",
    "material",
    "math",
    "matrix",
    "matter",
    "maximum",
    "maze",
    "meadow",
    "mean",
    "measure",
    "meat",
    "mechanic",
    "medal",
    "media",
    "melody",
    "melt",
    "member",
    "memory",
    "mention",
    "menu",
    "mercy",
    "merge",
    "merit",
    "merry",
    "mesh",
    "message",
    "metal",
    "method",
    "middle",
    "midnight",
    "milk",
    "million",
    "mimic",
    "mind",
    "minimum",
    "minor",
    "minute",
    "miracle",
    "mirror",
    "misery",
    "miss",
    "mistake",
    "mix",
    "mixed",
    "mixture",
    "mobile",
    "model",
    "modify",
    "mom",
    "moment",
    "monitor",
    "monkey",
    "monster",
    "month",
    "moon",
    "moral",
    "more",
    "morning",
    "mosquito",
    "mother",
    "motion",
    "motor",
    "mountain",
    "mouse",
    "move",
    "movie",
    "much",
    "muffin",
    "mule",
    "multiply",
    "muscle",
    "museum",
    "mushroom",
    "music",
    "must",
    "mutual",
    "myself",
    "mystery",
    "myth",
    "naive",
    "name",
    "napkin",
    "narrow",
    "nasty",
    "nation",
    "nature",
    "near",
    "neck",
    "need",
    "negative",
    "neglect",
    "neither",
    "nephew",
    "nerve",
    "nest",
    "net",
    "network",
    "neutral",
    "never",
    "news",
    "next",
    "nice",
    "night",
    "noble",
    "noise",
    "nominee",
    "noodle",
    "normal",
    "north",
    "nose",
    "notable",
    "note",
    "nothing",
    "notice",
    "novel",
    "now",
    "nuclear",
    "number",
    "nurse",
    "nut",
    "oak",
    "obey",
    "object",
    "oblige",
    "obscure",
    "observe",
    "obtain",
    "obvious",
    "occur",
    "ocean",
    "october",
    "odor",
    "off",
    "offer",
    "office",
    "often",
    "oil",
    "okay",
    "old",
    "olive",
    "olympic",
    "omit",
    "once",
    "one",
    "onion",
    "online",
    "only",
    "open",
    "opera",
    "opinion",
    "oppose",
    "option",
    "orange",
    "orbit",
    "orchard",
    "order",
    "ordinary",
    "organ",
    "orient",
    "original",
    "orphan",
    "ostrich",
    "other",
    "outdoor",
    "outer",
    "output",
    "outside",
    "oval",
    "oven",
    "over",
    "own",
    "owner",
    "oxygen",
    "oyster",
    "ozone",
    "pact",
    "paddle",
    "page",
    "pair",
    "palace",
    "palm",
    "panda",
    "panel",
    "panic",
    "panther",
    "paper",
    "parade",
    "parent",
    "park",
    "parrot",
    "party",
    "pass",
    "patch",
    "path",
    "patient",
    "patrol",
    "pattern",
    "pause",
    "pave",
    "payment",
    "peace",
    "peanut",
    "pear",
    "peasant",
    "pelican",
    "pen",
    "penalty",
    "pencil",
    "people",
    "pepper",
    "perfect",
    "permit",
    "person",
    "pet",
    "phone",
    "photo",
    "phrase",
    "physical",
    "piano",
    "picnic",
    "picture",
    "piece",
    "pig",
    "pigeon",
    "pill",
    "pilot",
    "pink",
    "pioneer",
    "pipe",
    "pistol",
    "pitch",
    "pizza",
    "place",
    "planet",
    "plastic",
    "plate",
    "play",
    "please",
    "pledge",
    "pluck",
    "plug",
    "plunge",
    "poem",
    "poet",
    "point",
    "polar",
    "pole",
    "police",
    "pond",
    "pony",
    "pool",
    "popular",
    "portion",
    "position",
    "possible",
    "post",
    "potato",
    "pottery",
    "poverty",
    "powder",
    "power",
    "practice",
    "praise",
    "predict",
    "prefer",
    "prepare",
    "present",
    "pretty",
    "prevent",
    "price",
    "pride",
    "primary",
    "print",
    "priority",
    "prison",
    "private",
    "prize",
    "problem",
    "process",
    "produce",
    "profit",
    "program",
    "project",
    "promote",
    "proof",
    "property",
    "prosper",
    "protect",
    "proud",
    "provide",
    "public",
    "pudding",
    "pull",
    "pulp",
    "pulse",
    "pumpkin",
    "punch",
    "pupil",
    "puppy",
    "purchase",
    "purity",
    "purpose",
    "purse",
    "push",
    "put",
    "puzzle",
    "pyramid",
    "quality",
    "quantum",
    "quarter",
    "question",
    "quick",
    "quit",
    "quiz",
    "quote",
    "rabbit",
    "raccoon",
    "race",
    "rack",
    "radar",
    "radio",
    "rail",
    "rain",
    "raise",
    "rally",
    "ramp",
    "ranch",
    "random",
    "range",
    "rapid",
    "rare",
    "rate",
    "rather",
    "raven",
    "raw",
    "razor",
    "ready",
    "real",
    "reason",
    "rebel",
    "rebuild",
    "recall",
    "receive",
    "recipe",
    "record",
    "recycle",
    "reduce",
    "reflect",
    "reform",
    "refuse",
    "region",
    "regret",
    "regular",
    "reject",
    "relax",
    "release",
    "relief",
    "rely",
    "remain",
    "remember",
    "remind",
    "remove",
    "render",
    "renew",
    "rent",
    "reopen",
    "repair",
    "repeat",
    "replace",
    "report",
    "require",
    "rescue",
    "resemble",
    "resist",
    "resource",
    "response",
    "result",
    "retire",
    "retreat",
    "return",
    "reunion",
    "reveal",
    "review",
    "reward",
    "rhythm",
    "rib",
    "ribbon",
    "rice",
    "rich",
    "ride",
    "ridge",
    "rifle",
    "right",
    "rigid",
    "ring",
    "riot",
    "ripple",
    "risk",
    "ritual",
    "rival",
    "river",
    "road",
    "roast",
    "robot",
    "robust",
    "rocket",
    "romance",
    "roof",
    "rookie",
    "room",
    "rose",
    "rotate",
    "rough",
    "round",
    "route",
    "royal",
    "rubber",
    "rude",
    "rug",
    "rule",
    "run",
    "runway",
    "rural",
    "sad",
    "saddle",
    "sadness",
    "safe",
    "sail",
    "salad",
    "salmon",
    "salon",
    "salt",
    "salute",
    "same",
    "sample",
    "sand",
    "satisfy",
    "satoshi",
    "sauce",
    "sausage",
    "save",
    "say",
    "scale",
    "scan",
    "scare",
    "scatter",
    "scene",
    "scheme",
    "school",
    "science",
    "scissors",
    "scorpion",
    "scout",
    "scrap",
    "screen",
    "script",
    "scrub",
    "sea",
    "search",
    "season",
    "seat",
    "second",
    "secret",
    "section",
    "security",
    "seed",
    "seek",
    "segment",
    "select",
    "sell",
    "seminar",
    "senior",
    "sense",
    "sentence",
    "series",
    "service",
    "session",
    "settle",
    "setup",
    "seven",
    "shadow",
    "shaft",
    "shallow",
    "share",
    "shed",
    "shell",
    "sheriff",
    "shield",
    "shift",
    "shine",
    "ship",
    "shiver",
    "shock",
    "shoe",
    "shoot",
    "shop",
    "short",
    "shoulder",
    "shove",
    "shrimp",
    "shrug",
    "shuffle",
    "shy",
    "sibling",
    "sick",
    "side",
    "siege",
    "sight",
    "sign",
    "silent",
    "silk",
    "silly",
    "silver",
    "similar",
    "simple",
    "since",
    "sing",
    "siren",
    "sister",
    "situate",
    "six",
    "size",
    "skate",
    "sketch",
    "ski",
    "skill",
    "skin",
    "skirt",
    "skull",
    "slab",
    "slam",
    "sleep",
    "slender",
    "slice",
    "slide",
    "slight",
    "slim",
    "slogan",
    "slot",
    "slow",
    "slush",
    "small",
    "smart",
    "smile",
    "smoke",
    "smooth",
    "snack",
    "snake",
    "snap",
    "sniff",
    "snow",
    "soap",
    "soccer",
    "social",
    "sock",
    "soda",
    "soft",
    "solar",
    "soldier",
    "solid",
    "solution",
    "solve",
    "someone",
    "song",
    "soon",
    "sorry",
    "sort",
    "soul",
    "sound",
    "soup",
    "source",
    "south",
    "space",
    "spare",
    "spatial",
    "spawn",
    "speak",
    "special",
    "speed",
    "spell",
    "spend",
    "sphere",
    "spice",
    "spider",
    "spike",
    "spin",
    "spirit",
    "split",
    "spoil",
    "sponsor",
    "spoon",
    "sport",
    "spot",
    "spray",
    "spread",
    "spring",
    "spy",
    "square",
    "squeeze",
    "squirrel",
    "stable",
    "stadium",
    "staff",
    "stage",
    "stairs",
    "stamp",
    "stand",
    "start",
    "state",
    "stay",
    "steak",
    "steel",
    "stem",
    "step",
    "stereo",
    "stick",
    "still",
    "sting",
    "stock",
    "stomach",
    "stone",
    "stool",
    "story",
    "stove",
    "strategy",
    "street",
    "strike",
    "strong",
    "struggle",
    "student",
    "stuff",
    "stumble",
    "style",
    "subject",
    "submit",
    "subway",
    "success",
    "such",
    "sudden",
    "suffer",
    "sugar",
    "suggest",
    "suit",
    "summer",
    "sun",
    "sunny",
    "sunset",
    "super",
    "supply",
    "supreme",
    "sure",
    "surface",
    "surge",
    "surprise",
    "surround",
    "survey",
    "suspect",
    "sustain",
    "swallow",
    "swamp",
    "swap",
    "swarm",
    "swear",
    "sweet",
    "swift",
    "swim",
    "swing",
    "switch",
    "sword",
    "symbol",
    "symptom",
    "syrup",
    "system",
    "table",
    "tackle",
    "tag",
    "tail",
    "talent",
    "talk",
    "tank",
    "tape",
    "target",
    "task",
    "taste",
    "tattoo",
    "taxi",
    "teach",
    "team",
    "tell",
    "ten",
    "tenant",
    "tennis",
    "tent",
    "term",
    "test",
    "text",
    "thank",
    "that",
    "theme",
    "then",
    "theory",
    "there",
    "they",
    "thing",
    "this",
    "thought",
    "three",
    "thrive",
    "throw",
    "thumb",
    "thunder",
    "ticket",
    "tide",
    "tiger",
    "tilt",
    "timber",
    "time",
    "tiny",
    "tip",
    "tired",
    "tissue",
    "title",
    "toast",
    "tobacco",
    "today",
    "toddler",
    "toe",
    "together",
    "toilet",
    "token",
    "tomato",
    "tomorrow",
    "tone",
    "tongue",
    "tonight",
    "tool",
    "tooth",
    "top",
    "topic",
    "topple",
    "torch",
    "tornado",
    "tortoise",
    "toss",
    "total",
    "tourist",
    "toward",
    "tower",
    "town",
    "toy",
    "track",
    "trade",
    "traffic",
    "tragic",
    "train",
    "transfer",
    "trap",
    "trash",
    "travel",
    "tray",
    "treat",
    "tree",
    "trend",
    "trial",
    "tribe",
    "trick",
    "trigger",
    "trim",
    "trip",
    "trophy",
    "trouble",
    "truck",
    "true",
    "truly",
    "trumpet",
    "trust",
    "truth",
    "try",
    "tube",
    "tuition",
    "tumble",
    "tuna",
    "tunnel",
    "turkey",
    "turn",
    "turtle",
    "twelve",
    "twenty",
    "twice",
    "twin",
    "twist",
    "two",
    "type",
    "typical",
    "ugly",
    "umbrella",
    "unable",
    "unaware",
    "uncle",
    "uncover",
    "under",
    "undo",
    "unfair",
    "unfold",
    "unhappy",
    "uniform",
    "unique",
    "unit",
    "universe",
    "unknown",
    "unlock",
    "until",
    "unusual",
    "unveil",
    "update",
    "upgrade",
    "uphold",
    "upon",
    "upper",
    "upset",
    "urban",
    "urge",
    "usage",
    "use",
    "used",
    "useful",
    "useless",
    "usual",
    "utility",
    "vacant",
    "vacuum",
    "vague",
    "valid",
    "valley",
    "valve",
    "van",
    "vanish",
    "vapor",
    "various",
    "vast",
    "vault",
    "vehicle",
    "velvet",
    "vendor",
    "venture",
    "venue",
    "verb",
    "verify",
    "version",
    "very",
    "vessel",
    "veteran",
    "viable",
    "vibrant",
    "vicious",
    "victory",
    "video",
    "view",
    "village",
    "vintage",
    "violin",
    "virtual",
    "virus",
    "visa",
    "visit",
    "visual",
    "vital",
    "vivid",
    "vocal",
    "voice",
    "void",
    "volcano",
    "volume",
    "vote",
    "voyage",
    "wage",
    "wagon",
    "wait",
    "walk",
    "wall",
    "walnut",
    "want",
    "warfare",
    "warm",
    "warrior",
    "wash",
    "wasp",
    "waste",
    "water",
    "wave",
    "way",
    "wealth",
    "weapon",
    "wear",
    "weasel",
    "weather",
    "web",
    "wedding",
    "weekend",
    "weird",
    "welcome",
    "west",
    "wet",
    "whale",
    "what",
    "wheat",
    "wheel",
    "when",
    "where",
    "whip",
    "whisper",
    "wide",
    "width",
    "wife",
    "wild",
    "will",
    "win",
    "window",
    "wine",
    "wing",
    "wink",
    "winner",
    "winter",
    "wire",
    "wisdom",
    "wise",
    "wish",
    "witness",
    "wolf",
    "woman",
    "wonder",
    "wood",
    "wool",
    "word",
    "work",
    "world",
    "worry",
    "worth",
    "wrap",
    "wreck",
    "wrestle",
    "wrist",
    "write",
    "wrong",
    "yard",
    "year",
    "yellow",
    "you",
    "young",
    "youth",
    "zebra",
    "zero",
    "zone",
    "zoo",
]



import decimal
import json
from decimal import localcontext
from typing import Optional, Union

import aiohttp
import pkg_resources
from aiohttp import ClientResponse
from coincurve import PrivateKey as CoincurvePrivateKey
from eth_abi import encode
from tronpy import Tron
from tronpy.keys import PrivateKey

from aiotx.clients._base_client import BlockMonitor
from aiotx.clients._evm_base_client import AioTxEVMBaseClient
from aiotx.exceptions import (
    CreateTransactionError,
    InvalidArgumentError,
    RpcConnectionError,
)
from aiotx.log import logger
from aiotx.types import BlockParam

units = {
    "sun": 1,
    "trx": 10**6,
}

MIN_SUN = 1
MAX_SUN = 10**18


class AioTxTRONClient(AioTxEVMBaseClient):
    def __init__(
        self,
        node_url: str,
        headers: dict = {},
    ):
        super().__init__(node_url, headers)
        self.monitor = TronMonitor(self)
        self._monitoring_task = None
        trc20_abi_json = pkg_resources.resource_string("aiotx.utils", "trc20_abi.json")
        self._trc20_abi = json.loads(trc20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._trc20_abi]

    def generate_address(self):
        client = Tron()
        return client.generate_address()

    def get_address_from_private_key(self, private_key: str):
        client = Tron()
        try:
            priv_key = PrivateKey(bytes.fromhex(private_key))
        except ValueError as e:
            raise ValueError(
                f"An error has been occurred during private key processing: {e}"
            )

        return client.generate_address(priv_key)

    def hex_address_to_base58(self, hex_address: str) -> str:
        # HACK sometimes we have address with 0x prefix?
        # Should we handle it somehow?
        if hex_address.startswith("0x"):
            hex_address = hex_address.replace("0x", "41")
        client = Tron()
        if not client.is_hex_address(hex_address):
            raise TypeError("Please provide hex address")
        return client.to_base58check_address(hex_address)

    def base58_to_hex_address(self, address) -> str:
        client = Tron()
        if not client.is_base58check_address(address):
            raise TypeError("Please provide base58 address")
        return client.to_hex_address(address)

    async def send(
        self,
        private_key: str,
        to_address: str,
        amount: int,
        memo: str = "",
    ) -> str:
        if not isinstance(memo, str):
            raise TypeError("Memo should be represented as a string!")
        sender_address_data = self.get_address_from_private_key(private_key)
        sender_address = sender_address_data["base58check_address"]
        created_txd = await self._create_transaction(
            sender_address, to_address, amount, memo
        )
        sig = self.sign_msg_hash(private_key, bytes.fromhex(created_txd["txID"]))
        result = await self.broadcast_transaction(
            [sig],
            created_txd["raw_data_hex"],
            created_txd["raw_data"],
            tx_id=created_txd["txID"],
        )
        return result["txid"]

    async def send_token(
        self,
        private_key: str,
        to_address: str,
        contract: str,
        amount: int,
        memo: str = "",
    ):
        if not isinstance(memo, str):
            raise TypeError("Memo should be represented as a string!")
        sender_address_data = self.get_address_from_private_key(private_key)
        sender_address = sender_address_data["base58check_address"]
        created_txd = await self._create_trc20_transfer_transaction(
            sender_address, to_address, amount, contract, memo=memo
        )
        tx_id = created_txd["txID"]
        sig = self.sign_msg_hash(private_key, bytes.fromhex(tx_id))
        result = await self.broadcast_transaction(
            [sig], created_txd["raw_data_hex"], created_txd["raw_data"], tx_id
        )
        return result["txid"]

    def sign_msg_hash(self, priv_key: str, message_hash: bytes) -> str:
        """Sign a message hash(sha256)."""
        private_key_bytes = bytes.fromhex(priv_key)
        signature_bytes = CoincurvePrivateKey(private_key_bytes).sign_recoverable(
            message_hash,
            hasher=None,
        )
        return signature_bytes.hex()

    async def broadcast_transaction(
        self,
        signature: list[str],
        raw_data_hex: str,
        raw_data: dict,
        tx_id: str,
        visible: bool = True,
    ):
        result = await self._make_api_call(
            {
                "signature": signature,
                "raw_data_hex": raw_data_hex,
                "raw_data": raw_data,
                "visible": visible,
                "txID": tx_id,
            },
            "POST",
            path="/wallet/broadcasttransaction",
        )
        return result

    async def _create_transaction(
        self, from_address, to_address, amount, memo: str = None
    ):
        payload = {
            "owner_address": from_address,
            "to_address": to_address,
            "amount": amount,
            "visible": True,
        }
        if memo is not None:
            payload["extra_data"] = memo.encode().hex()
        transaction = await self._make_api_call(
            payload, "POST", "/wallet/createtransaction"
        )
        if transaction.get("Error") is not None:
            raise CreateTransactionError(transaction.get("Error"))
        return transaction

    async def _create_trc20_transfer_transaction(
        self,
        sender_address: str,
        to_address: str,
        amount: int,
        contract_address: str,
        fee_limit: int = 15000000000,
        call_value: int = 0,
        visible: bool = True,
        memo: str = None,
    ) -> dict:
        # Construct the TRC20 token transfer transaction
        hex_eth_like_address = self.base58_to_hex_address(to_address).replace(
            "41", "0x"
        )

        transfer_data = encode(["address", "uint256"], [hex_eth_like_address, amount])
        parameter = transfer_data.hex()

        transaction = {
            "owner_address": sender_address,
            "contract_address": contract_address,
            "function_selector": "transfer(address,uint256)",
            "parameter": parameter,
            "fee_limit": fee_limit,
            "call_value": call_value,
            "visible": visible,
        }
        if memo is not None:
            transaction["extra_data"] = memo.encode().hex()

        # Make the API call to create the transaction
        result = await self._make_api_call(
            transaction, "POST", path="/wallet/triggersmartcontract"
        )
        if result.get("transaction") is None:
            raise CreateTransactionError(
                f"{result['result'].get('code')} {result['result'].get('message')}"
            )
        return result["transaction"]

    def to_sun(
        self, number: Union[int, float, str, decimal.Decimal], unit: str = "trx"
    ) -> int:
        """
        Takes a number of a unit and converts it to Sun (the smallest unit of TRX).
        """
        unit = unit.lower()
        if unit not in units:
            raise ValueError(f"Unknown unit. Must be one of {'/'.join(units.keys())}")

        if isinstance(number, int) or isinstance(number, str):
            d_number = decimal.Decimal(value=number)
        elif isinstance(number, float):
            d_number = decimal.Decimal(value=str(number))
        elif isinstance(number, decimal.Decimal):
            d_number = number
        else:
            raise TypeError(
                "Unsupported type. Must be one of integer, float, or string"
            )

        s_number = str(number)
        unit_value = units[unit]

        if d_number == decimal.Decimal(0):
            return 0

        if d_number < 1 and "." in s_number:
            with localcontext() as ctx:
                multiplier = len(s_number) - s_number.index(".") - 1
                ctx.prec = multiplier
                d_number = decimal.Decimal(value=number, context=ctx) * 10**multiplier
                unit_value /= 10**multiplier

        with localcontext() as ctx:
            ctx.prec = 999
            result_value = decimal.Decimal(value=d_number, context=ctx) * unit_value

        if result_value < MIN_SUN or result_value > MAX_SUN:
            raise ValueError("Resulting Sun value must be between 1 and 10^18")

        return int(result_value)

    def from_sun(self, number: Union[int, str], unit: str = "trx") -> decimal.Decimal:
        """
        Converts a value in Sun (the smallest unit of TRX) to the specified unit.
        """
        if isinstance(number, str):
            if number.startswith("0x"):
                number = int(number, 16)
            else:
                number = int(number)

        unit = unit.lower()
        if unit not in units:
            raise ValueError(f"Unknown unit. Must be one of {'/'.join(units.keys())}")

        unit_value = units[unit]
        result_value = decimal.Decimal(number) / unit_value

        return result_value

    async def get_balance(
        self, address, block_parameter: BlockParam = BlockParam.LATEST
    ) -> int:
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        return await super().get_balance(address, block_parameter)

    async def get_contract_balance(
        self, address, contract_address, block_parameter: BlockParam = BlockParam.LATEST
    ) -> int:
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        if client.is_base58check_address(contract_address):
            contract_address = self.base58_to_hex_address(contract_address)
        return await super().get_contract_balance(
            address, contract_address, block_parameter
        )

    async def get_contract_decimals(self, address: str):
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        return await super().get_contract_decimals(address)

    async def _make_api_call(self, payload, method, path) -> dict:
        url = self.node_url + path
        logger.info(f"api call payload: {payload} method: {method} path: {path}")
        async with aiohttp.ClientSession() as session:
            if method == "POST":
                headers = {"Content-Type": "application/json"}
                headers.update(self._headers)
                payload_json = json.dumps(payload)
                async with session.post(
                    url, data=payload_json, headers=headers
                ) as response:
                    return await self._process_api_answer(response)
            async with session.get(url) as response:
                return await self._process_api_answer(response)

    async def _process_api_answer(self, response: ClientResponse) -> dict:
        response_text = await response.text()
        logger.info(f"api call result: {response_text} status: {response.status}")
        if response.status != 200:
            raise RpcConnectionError(
                f"Node response status code: {response.status} response test: {response_text}"
            )
        result = await response.json()
        return result

    async def _make_rpc_call(self, payload, path="/jsonrpc") -> dict:
        payload["jsonrpc"] = "2.0"
        payload["id"] = 1
        payload_json = json.dumps(payload)
        logger.info(f"rpc call payload: {payload}")
        headers = {"Content-Type": "application/json"}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.node_url + path, data=payload_json, headers=headers
            ) as response:
                response_text = await response.text()
                logger.info(f"rpc call result: {response_text}")
                if response.status != 200:
                    raise RpcConnectionError(
                        f"Node response status code: {response.status} response test: {response_text}"
                    )
                result = await response.json()
                if "error" not in result.keys():
                    return result["result"]
                error_code = result["error"]["code"]
                error_message = result["error"]["message"]
                if (
                    "invalid characters encountered in Hex string" in error_message
                    or "invalid hash value" in error_message
                    or "invalid address hash value" in error_message
                ):
                    raise InvalidArgumentError(error_message)
                else:
                    raise RpcConnectionError(f"Error {error_code}: {error_message}")


class TronMonitor(BlockMonitor):
    def __init__(self, client: AioTxTRONClient, last_block: Optional[int] = None):
        self.client = client
        self.block_handlers = []
        self.transaction_handlers = []
        self.running = False
        self._last_block = last_block

    async def poll_blocks(self, _: int):
        network_last_block = await self.client.get_last_block_number()
        target_block = (
            network_last_block if self._latest_block is None else self._latest_block
        )
        if target_block > network_last_block:
            return
        block_data = await self.client.get_block_by_number(target_block)
        await self.process_transactions(block_data["transactions"])
        await self.process_block(target_block)
        self._latest_block = target_block + 1

    async def process_block(self, block):
        for handler in self.block_handlers:
            await handler(block)

    async def process_transactions(self, transactions):
        for transaction in transactions:
            transaction["aiotx_decoded_input"] = self.client.decode_transaction_input(
                transaction["input"]
            )
            for handler in self.transaction_handlers:
                await handler(transaction)



import os

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxETHClient
from aiotx.exceptions import (
    AioTxError,
    InvalidArgumentError,
    ReplacementTransactionUnderpriced,
    TransactionNotFound,
    WrongPrivateKey,
)

POLYGON_WALLET_PRIVATE_KEY = os.environ.get("TEST_POLYGON_WALLET_PRIVATE_KEY")
assert (
    POLYGON_WALLET_PRIVATE_KEY is not None
), "Please provide TEST_POLYGON_WALLET_PRIVATE_KEY"
CONTRACT = "0x0fd9e8d3af1aaee056eb9e802c3a762a667b1904"
DESTINATION_ADDRESS = "0x3ffeCb152F95f7122990ab16Eff4B0B5d533497e"


@vcr_c.use_cassette("polygon/get_last_block.yaml")
async def test_get_last_block(polygon_client: AioTxETHClient):
    block_id = await polygon_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("polygon/get_chain_id.yaml")
async def test_get_chain_id(polygon_client: AioTxETHClient):
    chain_id = await polygon_client.get_chain_id()
    assert isinstance(chain_id, int)
    assert chain_id == 80002


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_balance",
    [
        ("0x56ebc43d764761bc09d5918787672e2c8d46da5f", None, 0),
        ("0xf663792Be0EdD00AFFB8BBe4Ac6d8185efD5671d", None, 25472000000000000),
        ("0xA869aF2883DD6C881bC59594aDfB3e8b136321b4", None, 120000000000000),
        ("0xE27017DD9EF7e3D0A8321C2041Fc4402e89945b6", None, 130000000000000),
        ("0x39832298befa06a6db0c920ada31cfebAD2e33aB", None, 0),
        ("0xC4BfcCb1668d6e464f33a76bAdD8c8d7d341E04B", None, 0),
        ("0x68EfbC84d1Eabc193979beab2E2DDc20B219A14", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("polygon/get_balance.yaml")
async def test_get_balance(
    polygon_client: AioTxETHClient, wallet_address, expected_exception, expected_balance
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.get_balance(wallet_address)
    else:
        balance = await polygon_client.get_balance(wallet_address)
        assert isinstance(balance, int)
        assert balance == expected_balance


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        ("0x2860ef29c17bf8fd87a8ef4fcd64a76402b70011cba54f7e8ef2eecedf72a08f", None),
        ("0x1913ae3314b38254c07538910ebdf2e3a3d6562b084769ef6614a0f0f4e01367", None),
        ("0x711f485342d60846117686998618a23d148506b08a8f11462c6683a6f37b4845", None),
        ("0xa7c8f80df5559b87a46f881fba25d282f4bb15645211d43ec8531abe75b13cf1", None),
        ("0xea9bd00efa11d22593f3e636f00f8a1a36b836df3f7b4f1d945c1d7bc823bec5", None),
        ("0x75d284a35f58e1a847d5112e157e7bd41fab09feeac63df2f92048c863006cb2", None),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b73",
            TransactionNotFound,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b2",
            InvalidArgumentError,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf701b72",
            InvalidArgumentError,
        ),
    ],
)
@vcr_c.use_cassette("polygon/get_transaction.yaml")
async def test_get_transaction(
    polygon_client: AioTxETHClient, tx_id, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.get_transaction(tx_id)
    else:
        tx = await polygon_client.get_transaction(tx_id)
        assert "aiotx_decoded_input" in tx.keys()


@pytest.mark.parametrize(
    "wallet_address, contract, expected_exception, expected_balance",
    [
        (
            "0x3ce3668dfccaee1a348166b2df1c88f0d6a1d2db",
            CONTRACT,
            None,
            38000000000000000000,
        ),
        (
            "0xe68db62564964df789e00927f86b5d6572f77634",
            CONTRACT,
            None,
            25000000000000000000,
        ),
        (
            "0x15331fe7e3638c6edd5833286c7fdceb3506d46c",
            CONTRACT,
            None,
            182000000000000000000,
        ),
        (
            "0x3ce3668dfccaee1a348166b2df1c88f0d6a1d2db",
            CONTRACT,
            None,
            38000000000000000000,
        ),
        (
            "0x94524d4d3b8ad325f48d637b65f83e50165c6bbb",
            CONTRACT,
            None,
            25000000000000000000,
        ),
        (
            "0x440eb4c94b0aaa8bb3e154fa1361adef71f606e0",
            CONTRACT,
            None,
            125000000000000000000,
        ),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", CONTRACT, None, 0),
        (
            "0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C",
            "Eeaa7E347ee4557f9a6F71C",
            InvalidArgumentError,
            0,
        ),
    ],
)
@vcr_c.use_cassette("polygon/get_token_balance.yaml")
async def test_get_token_balance(
    polygon_client: AioTxETHClient,
    wallet_address,
    contract,
    expected_exception,
    expected_balance,
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.get_contract_balance(wallet_address, contract)
    else:
        balance = await polygon_client.get_contract_balance(wallet_address, contract)
        assert isinstance(balance, int)
        assert expected_balance == balance


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_count",
    [
        ("0x3ce3668dfccaee1a348166b2df1c88f0d6a1d2db", None, 19),
        ("0xe68db62564964df789e00927f86b5d6572f77634", None, 0),
        ("0x15331fe7e3638c6edd5833286c7fdceb3506d46c", None, 2),
        ("0x94524d4d3b8ad325f48d637b65f83e50165c6bbb", None, 31),
        ("0x440eb4c94b0aaa8bb3e154fa1361adef71f606e0", None, 4),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", None, 0),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71", InvalidArgumentError, 0),
        ("DA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("polygon/get_transaction_count.yaml")
async def test_get_transaction_count(
    polygon_client: AioTxETHClient, wallet_address, expected_exception, expected_count
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.get_transactions_count(wallet_address)
    else:
        count = await polygon_client.get_transactions_count(wallet_address)
        assert isinstance(count, int)
        assert expected_count == count


@vcr_c.use_cassette("polygon/get_gas_price.yaml")
async def test_get_gas_price(polygon_client: AioTxETHClient):
    result = await polygon_client.get_gas_price()
    assert isinstance(result, int)


@pytest.mark.parametrize(
    "private_key, to_address, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5",
            0.00001,
            5,
            21000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            21000,
            None,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "0xf9E35E4e1CbcF08E84d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            21000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            ReplacementTransactionUnderpriced,
        ),
        (POLYGON_WALLET_PRIVATE_KEY, DESTINATION_ADDRESS, 5, 5, 21000, AioTxError),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            0.00001,
            0,
            21000,
            AioTxError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            0,
            5,
            21000,
            ReplacementTransactionUnderpriced,
        ),
        (POLYGON_WALLET_PRIVATE_KEY, DESTINATION_ADDRESS, 0.00001, 5, 0, AioTxError),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            0.00001,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("polygon/send_transaction.yaml")
async def test_send_transaction(
    polygon_client: AioTxETHClient,
    private_key,
    to_address,
    amount,
    gas_price,
    gas_limit,
    expected_exception,
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = polygon_client.to_wei(gas_price, "gwei")
    wei_amount = polygon_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.send(
                private_key,
                to_address,
                wei_amount,
                gas_price=gas_price,
                gas_limit=gas_limit,
            )
    else:
        result = await polygon_client.send(
            private_key,
            to_address,
            wei_amount,
            gas_price=gas_price,
            gas_limit=gas_limit,
        )
        assert isinstance(result, str)


@pytest.mark.parametrize(
    "private_key, to_address, amount, expected_exception",
    [
        (POLYGON_WALLET_PRIVATE_KEY, DESTINATION_ADDRESS, 0.00001, None),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            0.00001,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("polygon/send_transaction_with_auto_gas.yaml")
async def test_send_transaction_with_auto_gas(
    polygon_client: AioTxETHClient, private_key, to_address, amount, expected_exception
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    wei_amount = polygon_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.send(private_key, to_address, wei_amount)
    else:
        result = await polygon_client.send(private_key, to_address, wei_amount)
        assert isinstance(result, str)


@vcr_c.use_cassette("polygon/send_transaction_with_custom_nonce.yaml")
async def test_send_transaction_with_custom_nonce(polygon_client: AioTxETHClient):
    wei_amount = polygon_client.to_wei(0.00001, "ether")
    sender_address = polygon_client.get_address_from_private_key(
        POLYGON_WALLET_PRIVATE_KEY
    )
    nonce = await polygon_client.get_transactions_count(sender_address)
    first_tx = await polygon_client.send(
        POLYGON_WALLET_PRIVATE_KEY, DESTINATION_ADDRESS, wei_amount, nonce=nonce
    )
    assert isinstance(first_tx, str)
    second_tx = await polygon_client.send(
        POLYGON_WALLET_PRIVATE_KEY, sender_address, wei_amount, nonce=nonce + 1
    )
    assert isinstance(second_tx, str)


@pytest.mark.parametrize(
    "contract, expected_decimals, expected_exception",
    [
        (CONTRACT, 18, None),
        ("0x337610d27c682E347C9cD60BD4b3b107C9d34dD", 18, InvalidArgumentError),
        ("0x41E94Eb019C0762f9Bfcf9Fb1E58725BfB0e7582", 6, None),
    ],
)
@vcr_c.use_cassette("polygon/get_contract_decimals.yaml")
async def test_get_contract_decimals(
    polygon_client: AioTxETHClient, contract, expected_decimals, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.get_contract_decimals(contract)
    else:
        decimals = await polygon_client.get_contract_decimals(contract)
        assert decimals == expected_decimals


@vcr_c.use_cassette("polygon/send_token_transaction_with_auto_params.yaml")
async def test_send_transaction_with_auto_params(polygon_client: AioTxETHClient):
    wei_amount = polygon_client.to_wei(0.1, "ether")
    tx_id = await polygon_client.send_token(
        POLYGON_WALLET_PRIVATE_KEY, DESTINATION_ADDRESS, CONTRACT, wei_amount
    )
    assert isinstance(tx_id, str)


@pytest.mark.parametrize(
    "private_key, to_address, contract, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "f9E5E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b",
            CONTRACT,
            1,
            5,
            61000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            61000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            1,
            5,
            61000,
            None,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b",
            1,
            5,
            61000,
            ValueError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            0.3,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            1000,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            0,
            61000,
            AioTxError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            0,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            0,
            AioTxError,
        ),
        (
            POLYGON_WALLET_PRIVATE_KEY,
            DESTINATION_ADDRESS,
            CONTRACT,
            0.4,
            5,
            61000,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("polygon/send_token_transaction.yaml")
async def test_send_token_transaction(
    polygon_client: AioTxETHClient,
    private_key,
    to_address,
    contract,
    amount,
    gas_price,
    gas_limit,
    expected_exception,
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = polygon_client.to_wei(gas_price, "gwei")
    wei_amount = polygon_client.to_wei(amount, "gwei")

    if expected_exception:
        with pytest.raises(expected_exception):
            await polygon_client.send_token(
                private_key,
                to_address,
                contract,
                wei_amount,
                gas_price=gas_price,
                gas_limit=gas_limit,
            )
    else:
        result = await polygon_client.send_token(
            private_key,
            to_address,
            contract,
            wei_amount,
            gas_price=gas_price,
            gas_limit=gas_limit,
        )
        assert isinstance(result, str)


@vcr_c.use_cassette("polygon/send_token_transaction_with_custom_nonce.yaml")
async def test_send_token_transaction_with_custom_nonce(polygon_client: AioTxETHClient):
    wei_amount = polygon_client.to_wei(0.1, "gwei")
    sender_address = polygon_client.get_address_from_private_key(
        POLYGON_WALLET_PRIVATE_KEY
    )
    nonce = await polygon_client.get_transactions_count(sender_address)
    first_tx = await polygon_client.send_token(
        POLYGON_WALLET_PRIVATE_KEY,
        DESTINATION_ADDRESS,
        CONTRACT,
        wei_amount,
        nonce=nonce,
    )
    assert isinstance(first_tx, str)
    second_tx = await polygon_client.send_token(
        POLYGON_WALLET_PRIVATE_KEY,
        sender_address,
        CONTRACT,
        wei_amount,
        nonce=nonce + 1,
    )
    assert isinstance(second_tx, str)



import asyncio

from aiotx.clients import AioTxTONClient

ton_client = AioTxTONClient("https://testnet.toncenter.com/api/v2", workchain=0)
# We are adding workchain here because testnet.toncenter working bad and identify itself as -1 but it should be 0
# If you are using any other provider it should work fine without workchain param


# wallet = asyncio.run(ton_client.generate_address())
# print("wallet", wallet)

amount = ton_client.to_nano(0.0001)
mnemonic_str = "truck salt rib coffee fold tree album often rice nominee green minor lemon ritual gossip divorce monkey garment symbol super lonely liar pond coin"

# balance = asyncio.run(ton_client.get_balance("UQDU1hdX6SeHmrvzvyjIrLEWUAdJUJar2sw8haIuT_5n-FLh"))

tx_id = asyncio.run(
    ton_client.send(
        mnemonic_str,
        "0QAEhA1CupMp7uMOUfHHoh7sqAMNu1xQOydf8fQf-ATpkbpT",
        amount,
        memo="for Melanty with love",
    )
)
print(tx_id)



import asyncio
import json
from decimal import Decimal
from typing import Optional, Union

import aiohttp
from aiohttp.client_exceptions import ClientError, ClientOSError
from bitcoinlib.encoding import pubkeyhash_to_addr_bech32
from bitcoinlib.keys import HDKey, Key
from bitcoinlib.networks import Network
from bitcoinlib.transactions import Transaction
from sqlalchemy import BigInteger, Boolean, Column, Integer, String, select, update
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.pool import NullPool

from aiotx.clients._base_client import AioTxClient, BlockMonitor
from aiotx.exceptions import (
    AioTxError,
    BlockNotFoundError,
    InsufficientFunds,
    InternalJSONRPCError,
    InvalidArgumentError,
    InvalidRequestError,
    MethodNotFoundError,
    RpcConnectionError,
)
from aiotx.log import logger
from aiotx.types import FeeEstimate

Base = declarative_base()


def create_address_model(currency_name):
    class Address(Base):
        __tablename__ = f"{currency_name}_addresses"
        __table_args__ = {"extend_existing": True}

        address = Column(String(255), primary_key=True)
        block_number = Column(Integer)

    return Address


def create_utxo_model(currency_name):
    class UTXO(Base):
        __tablename__ = f"{currency_name}_utxo"
        __table_args__ = {"extend_existing": True}

        tx_id = Column(String(255), primary_key=True)
        output_n = Column(Integer, primary_key=True)
        address = Column(String(255))
        amount_satoshi = Column(BigInteger)
        used = Column(Boolean, default=False)

    return UTXO


def create_last_block_model(currency_name):
    class LastBlock(Base):
        __tablename__ = f"{currency_name}_last_block"
        __table_args__ = {"extend_existing": True}

        block_number = Column(Integer, primary_key=True)

    return LastBlock


class AioTxUTXOClient(AioTxClient):
    def __init__(
        self,
        node_url: str,
        headers,
        testnet: bool,
        node_username,
        node_password,
        network_name,
        db_url,
    ):
        super().__init__(node_url, headers)
        self.node_username = node_username
        self.node_password = node_password
        self.testnet = testnet
        self._network = Network(network_name)
        self.monitor = UTXOMonitor(self, db_url)
        asyncio.run(self.monitor._init_db())

    @staticmethod
    def to_satoshi(amount: Union[int, float]) -> int:
        return int(Decimal(str(amount)) * Decimal(10**8))

    @staticmethod
    def from_satoshi(amount: int) -> float:
        return amount / 10**8

    async def generate_address(self) -> dict:
        hdkey = HDKey()
        derivation_path = f"m/84'/{self._network.bip44_cointype}'/0'/0/0"
        private_key = hdkey.subkey_for_path(derivation_path).private_hex
        hash160 = hdkey.subkey_for_path(derivation_path).hash160
        address = pubkeyhash_to_addr_bech32(
            hash160, prefix=self._network.prefix_bech32, witver=0, separator="1"
        )
        last_block_number = await self.get_last_block_number()
        await self.import_address(address, last_block_number)
        return private_key, address

    def get_address_from_private_key(self, private_key):
        key = Key(private_key)
        address = pubkeyhash_to_addr_bech32(
            key.hash160, prefix=self._network.prefix_bech32, witver=0, separator="1"
        )
        return {
            "private_key": private_key,
            "public_key": key.public_hex,
            "address": address,
        }

    async def import_address(self, address: str, block_number: int = None):
        await self.monitor._add_new_address(address, block_number)

    async def get_last_block_number(self) -> int:
        payload = {"method": "getblockcount", "params": []}
        result = await self._make_rpc_call(payload)
        return result["result"]

    async def get_block_by_number(self, block_number: int, verbosity: int = 2):
        payload = {"method": "getblockhash", "params": [block_number]}
        block_hash = await self._make_rpc_call(payload)
        payload = {"method": "getblock", "params": [block_hash["result"], verbosity]}
        result = await self._make_rpc_call(payload)
        return result["result"]

    async def get_balance(self, address: str) -> int:
        utxo_data = await self.monitor._get_utxo_data(address)
        if len(utxo_data) == 0:
            return 0
        return sum(utxo.amount_satoshi for utxo in utxo_data)

    async def _build_and_send_transaction(
        self,
        private_key: str,
        destinations: dict[str, int],
        conf_target: int,
        estimate_mode: FeeEstimate,
        total_fee: Optional[int],
        fee_per_byte: Optional[int],
        deduct_fee: bool,
    ) -> str:
        from_wallet = self.get_address_from_private_key(private_key)
        from_address = from_wallet["address"]
        utxo_list = await self.monitor._get_utxo_data(from_address)

        if total_fee is None:
            empty_fee_transaction, _, _ = await self._create_transaction(
                destinations, utxo_list, from_address, 0, deduct_fee
            )
            if fee_per_byte is None:
                fee_per_kb = await self.estimate_smart_fee(conf_target, estimate_mode)
            else:
                fee_per_kb = fee_per_byte * 1024
            empty_fee_transaction.fee_per_kb = fee_per_kb
            empty_fee_transaction = self._sign_transaction(
                empty_fee_transaction, [private_key]
            )
            empty_fee_transaction.estimate_size()
            total_fee = empty_fee_transaction.calculate_fee()

        transaction, inputs_used, outputs_used = await self._create_transaction(
            destinations, utxo_list, from_address, total_fee, deduct_fee
        )

        signed_tx = self._sign_transaction(
            transaction, [private_key] * len(inputs_used)
        )
        txid = await self._send_transaction(signed_tx.raw_hex())
        await self._mark_inputs_as_used(inputs_used)
        await self._save_pending_outputs(outputs_used, from_address, txid)
        return txid

    async def _save_pending_outputs(self, outputs_used, from_address: str, tx_id: str):
        for i, output in enumerate(outputs_used):
            destination_address = output[0]
            if from_address != destination_address:
                continue
            amount = output[1]
            await self.monitor._add_new_utxo(from_address, tx_id, amount, i)

    async def send(
        self,
        private_key: str,
        to_address: str,
        amount: int,
        total_fee: int = None,
        fee_per_byte: int = None,
        conf_target: int = 6,
        estimate_mode: FeeEstimate = FeeEstimate.CONSERVATIVE,
        deduct_fee: bool = False,
    ) -> str:
        return await self._build_and_send_transaction(
            private_key,
            {to_address: amount},
            conf_target,
            estimate_mode,
            total_fee,
            fee_per_byte,
            deduct_fee,
        )

    async def send_bulk(
        self,
        private_key: str,
        destinations: dict[str, int],
        total_fee: int = None,
        fee_per_byte: int = None,
        conf_target: int = 6,
        estimate_mode: FeeEstimate = FeeEstimate.CONSERVATIVE,
        deduct_fee: bool = False,
    ) -> str:
        return await self._build_and_send_transaction(
            private_key,
            destinations,
            conf_target,
            estimate_mode,
            total_fee,
            fee_per_byte,
            deduct_fee,
        )

    async def _create_transaction(
        self,
        destinations: dict[str, int],
        utxo_list: list,
        from_address: str,
        fee: int,
        deduct_fee: bool,
    ) -> tuple[Transaction, list]:
        transaction = Transaction(network=self._network.name, witness_type="segwit")
        inputs = []
        outputs = []
        total_amount = sum(destinations.values())
        total_value = 0
        for utxo in utxo_list:
            input_data = (utxo.tx_id, utxo.output_n, utxo.amount_satoshi)
            inputs.append(input_data)
            total_value += utxo.amount_satoshi
            if total_value >= total_amount + fee:
                break
        total_spend = total_amount + fee if not deduct_fee else total_amount
        if total_value < total_spend:
            raise InsufficientFunds(
                f"We have only {total_value} satoshi and it's {total_amount + fee} at least needed to cover that transaction!"
            )

        if deduct_fee:
            leftover = total_value - total_amount
        else:
            leftover = int(total_value - total_amount - fee)

        if leftover > 0:
            outputs.append((from_address, leftover))

        deducted_fee_amount = 0 if not deduct_fee else fee / len(destinations)

        for address, amount in destinations.items():
            outputs.append((address, amount - deducted_fee_amount))

        for input_data in inputs:
            prev_tx_id, prev_out_index, value = input_data
            transaction.add_input(
                prev_txid=prev_tx_id,
                output_n=prev_out_index,
                value=value,
                witness_type="segwit",
            )

        for output_data in outputs:
            address, value = output_data
            transaction.add_output(value=value, address=address)

        return transaction, inputs, outputs

    def _sign_transaction(
        self, transaction: Transaction, private_keys: list[str]
    ) -> Transaction:
        for i, private_key in enumerate(private_keys):
            key = Key(private_key)
            transaction.sign(key, i)
        return transaction

    async def _mark_inputs_as_used(self, inputs: list):
        for input in inputs:
            await self.monitor._mark_utxo_used(input[0], input[1])

    async def _send_transaction(self, raw_transaction: str) -> str:
        payload = {"method": "sendrawtransaction", "params": [raw_transaction]}
        result = await self._make_rpc_call(payload)
        return result["result"]

    async def estimate_smart_fee(
        self,
        conf_target: int = 6,
        estimate_mode: FeeEstimate = FeeEstimate.CONSERVATIVE,
    ) -> int:
        payload = {
            "method": "estimatesmartfee",
            "params": [conf_target, estimate_mode.value],
        }
        result = await self._make_rpc_call(payload)
        return self.to_satoshi(result["result"]["feerate"])

    async def _make_rpc_call(self, payload) -> dict:
        payload["jsonrpc"] = "2.0"
        payload["id"] = "curltest"
        logger.info(f"rpc call payload: {payload}")
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.node_url,
                data=json.dumps(payload),
                headers=self._headers,
                auth=aiohttp.BasicAuth(self.node_username, self.node_password),
            ) as response:
                if response.status != 200:
                    raise RpcConnectionError(await response.text())
                result = await response.json()
                error = result.get("error")
                logger.info(f"rpc call result: {result}")
                if error is None:
                    return result

                error_code = error.get("code")
                error_message = error.get("message")
                if error_code == -5:
                    raise BlockNotFoundError(error_message)
                elif error_code == -8:
                    raise InvalidArgumentError(error_message)
                elif error_code == -32600:
                    raise InvalidRequestError(error_message)
                elif error_code == -32601:
                    raise MethodNotFoundError(error_message)
                elif error_code == -32603:
                    raise InternalJSONRPCError(error_message)
                else:
                    raise AioTxError(f"Error {error_code}: {error_message}")


class UTXOMonitor(BlockMonitor):
    def __init__(self, client: AioTxUTXOClient, db_url):
        self.client = client
        self.block_handlers = []
        self.transaction_handlers = []
        self.running = False
        self._db_url = db_url
        self._engine = create_async_engine(db_url, poolclass=NullPool)
        self._session = sessionmaker(
            self._engine, class_=AsyncSession, expire_on_commit=False
        )
        self.Address = create_address_model(self.client._network.name)
        self.UTXO = create_utxo_model(self.client._network.name)
        self.LastBlock = create_last_block_model(self.client._network.name)

    async def poll_blocks(self, _: int):
        network_last_block = await self.client.get_last_block_number()
        local_latest_block = await self._get_last_block()
        if local_latest_block is None:
            local_latest_block = network_last_block
        if network_last_block < local_latest_block:
            return
        block_data = await self.client.get_block_by_number(local_latest_block)
        await self.process_block(local_latest_block, block_data)
        next_block = local_latest_block + 1
        await self._update_last_block(next_block)

    async def process_block(self, block_number, block_data):
        await self._update_last_block(block_number)
        for handler in self.block_handlers:
            await handler(block_number)

        addresses = await self._get_addresses()
        for transaction in block_data["tx"]:
            for output in transaction["vout"]:
                outputs_scriptPubKey = output.get("scriptPubKey")
                if outputs_scriptPubKey is None:
                    continue

                output_address_list = outputs_scriptPubKey.get("addresses")
                output_address = outputs_scriptPubKey.get("address")

                if output_address is None and output_address_list is None:
                    continue

                if output_address_list is not None:
                    to_address = output_address_list[0]
                else:
                    to_address = output_address

                if to_address not in addresses:
                    continue

                value = self.client.to_satoshi(output["value"])
                output_n = output["n"]
                await self._add_new_utxo(
                    to_address, transaction["txid"], value, output_n
                )

        all_utxo_tx_ids = await self._get_all_utxo_tx_ids()
        for transaction in block_data["tx"]:
            for input_utxo in transaction["vin"]:
                txid = input_utxo.get("txid")
                vout = input_utxo.get("vout")
                if txid is None or vout is None:
                    continue
                if txid in all_utxo_tx_ids:
                    await self._process_input_utxo(txid, vout)

        for transaction in block_data["tx"]:
            for handler in self.transaction_handlers:
                await handler(transaction)

    async def _init_db(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
            try:
                last_known_block = await self.client.get_last_block_number()
            except (ClientError, ClientOSError) as e:
                logger.warning(
                    f"Can't get the last known block number during database initialization for the UTXO client. Please check your connection. {e}"
                )
            else:
                await self._init_last_block(last_known_block)

    async def _drop_tables(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.drop_all)

    async def _add_new_address(self, address: str, block_number: int = None):
        last_known_block = await self.client.get_last_block_number()
        if block_number is None:
            block_number = last_known_block

        async with self._session() as session:
            async with session.begin():
                existing_address = await session.scalar(
                    select(self.Address).filter_by(address=address)
                )
                if existing_address:
                    existing_address.block_number = block_number
                else:
                    session.add(
                        self.Address(address=address, block_number=block_number)
                    )
                await session.commit()
        last_known_block = await self._get_last_block()
        if last_known_block is None or last_known_block > block_number:
            await self._update_last_block(block_number)

    async def _add_new_utxo(
        self, address: str, tx_id: str, amount: int, output_n: int
    ) -> None:
        async with self._session() as session:
            async with session.begin():
                existing_utxo = await session.scalar(
                    select(self.UTXO).filter_by(
                        address=address, tx_id=tx_id, output_n=output_n
                    )
                )
                if existing_utxo:
                    existing_utxo.used = False
                else:
                    session.add(
                        self.UTXO(
                            address=address,
                            tx_id=tx_id,
                            amount_satoshi=amount,
                            output_n=output_n,
                        )
                    )
                await session.commit()

    async def _update_last_block(self, block_number: int) -> None:
        async with self._session() as session:
            async with session.begin():
                data = await session.execute(select(self.LastBlock))
                last_block = data.scalar()
                if last_block is None:
                    await self._init_last_block(block_number)
                else:
                    last_block.block_number = block_number
                    await session.commit()

    async def _init_last_block(self, block_number: int) -> None:
        async with self._session() as session:
            async with session.begin():
                data = await session.execute(select(self.LastBlock))
                last_block = data.scalar()
                if last_block is None:
                    session.add(self.LastBlock(block_number=block_number))
                await session.commit()

    async def _process_input_utxo(self, txid: str, vout: int) -> None:
        utxo = await self._get_utxo(txid, vout)
        if utxo:
            await self._delete_utxo(txid, vout)

    async def _get_utxo_data(self, address: str, spent=False):
        async with self._session() as session:
            result = await session.execute(
                select(
                    self.UTXO.tx_id,
                    self.UTXO.output_n,
                    self.UTXO.amount_satoshi,
                    self.UTXO.used,
                ).where((self.UTXO.address == address) & (self.UTXO.used == spent))
            )
            return result.fetchall()

    async def _mark_utxo_used(self, tx_id: str, output_n: str) -> None:
        async with self._session() as session:
            async with session.begin():
                stmt = (
                    update(self.UTXO)
                    .where(
                        (self.UTXO.tx_id == tx_id) & (self.UTXO.output_n == output_n)
                    )
                    .values(used=True)
                )
                await session.execute(stmt)
                await session.commit()

    async def _get_utxo(self, tx_id: str, output_n: int):
        async with self._session() as session:
            result = await session.execute(
                select(self.UTXO).where(
                    self.UTXO.tx_id == tx_id, self.UTXO.output_n == output_n
                )
            )
            return result.scalar()

    async def _get_all_utxo_tx_ids(self):
        """
        Function to optimize UTXO checking on new block to not make a SELECT for each input in the block
        """
        async with self._session() as session:
            result = await session.execute(select(self.UTXO.tx_id))
            return {row[0] for row in result.fetchall()}

    async def _delete_utxo(self, tx_id: str, output_n: int) -> None:
        async with self._session() as session:
            async with session.begin():
                await session.delete(await session.get(self.UTXO, (tx_id, output_n)))
                await session.commit()

    async def _get_last_block(self) -> Optional[int]:
        async with self._session() as session:
            result = await session.execute(select(self.LastBlock.block_number))
            return result.scalar()

    async def _get_addresses(self) -> list[str]:
        async with self._session() as session:
            result = await session.execute(select(self.Address.address))
            return {row[0] for row in result.fetchall()}



from math import floor
from typing import List, Tuple

from ....boc import Cell, DictBuilder
from ....utils import Address
from ... import Contract
from .nft_utils import create_offchain_uri_cell, serialize_uri


class NFTCollection(Contract):
    code = "B5EE9C724102140100021F000114FF00F4A413F4BCF2C80B0102016202030202CD04050201200E0F04E7D10638048ADF000E8698180B8D848ADF07D201800E98FE99FF6A2687D20699FEA6A6A184108349E9CA829405D47141BAF8280E8410854658056B84008646582A802E78B127D010A65B509E58FE59F80E78B64C0207D80701B28B9E382F970C892E000F18112E001718112E001F181181981E0024060708090201200A0B00603502D33F5313BBF2E1925313BA01FA00D43028103459F0068E1201A44343C85005CF1613CB3FCCCCCCC9ED54925F05E200A6357003D4308E378040F4966FA5208E2906A4208100FABE93F2C18FDE81019321A05325BBF2F402FA00D43022544B30F00623BA9302A402DE04926C21E2B3E6303250444313C85005CF1613CB3FCCCCCCC9ED54002C323401FA40304144C85005CF1613CB3FCCCCCCC9ED54003C8E15D4D43010344130C85005CF1613CB3FCCCCCCC9ED54E05F04840FF2F00201200C0D003D45AF0047021F005778018C8CB0558CF165004FA0213CB6B12CCCCC971FB008002D007232CFFE0A33C5B25C083232C044FD003D0032C03260001B3E401D3232C084B281F2FFF2742002012010110025BC82DF6A2687D20699FEA6A6A182DE86A182C40043B8B5D31ED44D0FA40D33FD4D4D43010245F04D0D431D430D071C8CB0701CF16CCC980201201213002FB5DAFDA89A1F481A67FA9A9A860D883A1A61FA61FF480610002DB4F47DA89A1F481A67FA9A9A86028BE09E008E003E00B01A500C6E"

    def __init__(self, **kwargs):
        self.code = kwargs.get("code") or self.code
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)
        self.options["royalty_base"] = self.options.get("royalty_base", 1000)
        self.options["royalty_factor"] = floor(
            self.options.get("royalty", 0) * self.options["royalty_base"]
        )

    def create_content_cell(self, params) -> Cell:
        collection_content_cell = create_offchain_uri_cell(
            params["collection_content_uri"]
        )
        common_content_cell = Cell()
        common_content_cell.bits.write_bytes(
            serialize_uri(params["nft_item_content_base_uri"])
        )
        content_cell = Cell()
        content_cell.refs.append(collection_content_cell)
        content_cell.refs.append(common_content_cell)
        return content_cell

    def create_royalty_cell(self, params) -> Cell:
        royalty_cell = Cell()
        royalty_cell.bits.write_uint(params["royalty_factor"], 16)
        royalty_cell.bits.write_uint(params["royalty_base"], 16)
        royalty_cell.bits.write_address(params["royalty_address"])
        return royalty_cell

    def create_data_cell(self) -> Cell:
        cell = Cell()
        cell.bits.write_address(self.options["owner_address"])
        cell.bits.write_uint(0, 64)  # next_item_index
        cell.refs.append(self.create_content_cell(self.options))
        cell.refs.append(Cell.one_from_boc(self.options["nft_item_code_hex"]))
        cell.refs.append(self.create_royalty_cell(self.options))
        return cell

    def create_mint_body(
        self,
        item_index: int,
        new_owner_address: Address,
        item_content_uri: str,
        amount: int = 50000000,
        query_id: int = 0,
    ) -> Cell:
        body = Cell()
        body.bits.write_uint(1, 32)
        body.bits.write_uint(query_id, 64)
        body.bits.write_uint(item_index, 64)
        body.bits.write_grams(amount)
        content_cell = Cell()
        content_cell.bits.write_address(new_owner_address)
        uri_content = Cell()
        uri_content.bits.write_bytes(serialize_uri(item_content_uri))
        content_cell.refs.append(uri_content)
        body.refs.append(content_cell)
        return body

    def create_batch_mint_body(
        self,
        from_item_index: int,
        contents_and_owners: List[Tuple[str, Address]],
        amount_per_one: int = 50000000,
        query_id: int = 0,
    ) -> Cell:
        body = Cell()
        body.bits.write_uint(2, 32)
        body.bits.write_uint(query_id, 64)
        deploy_list = DictBuilder(64)
        for i, (item_content_uri, new_owner_address) in enumerate(contents_and_owners):
            item = Cell()
            item.bits.write_grams(amount_per_one)
            content = Cell()
            content.bits.write_address(new_owner_address)
            uri_content = Cell()
            uri_content.bits.write_bytes(serialize_uri(item_content_uri))
            content.refs.append(uri_content)
            item.refs.append(content)
            deploy_list.store_cell(i + from_item_index, item)
        body.refs.append(deploy_list.end_dict())
        return body

    def create_get_royalty_params_body(self, query_id: int = 0) -> Cell:
        body = Cell()
        body.bits.write_uint(0x693D3950, 32)  # OP
        body.bits.write_uint(query_id, 64)  # query_id
        return body

    def create_change_owner_body(
        self, new_owner_address: Address, query_id: int = 0
    ) -> Cell:
        body = Cell()
        body.bits.write_uint(3, 32)  # OP
        body.bits.write_uint(query_id, 64)  # query_id
        body.bits.write_address(new_owner_address)
        return body

    def create_edit_content_body(self, params) -> Cell:
        if params["royalty"] > 1:
            raise Exception("royalty must be less than 1")

        body = Cell()
        body.bits.write_uint(4, 32)  # OP
        body.bits.write_uint(params.get("query_id", 0), 64)  # query_id
        body.refs.append(self.create_content_cell(params))
        body.refs.append(self.create_royalty_cell(params))
        return body



import asyncio

from aiotx.clients import AioTxBSCClient, AioTxETHClient

BSC_TEST_NODE_URL = "https://bsc-testnet-rpc.publicnode.com"
ETH_TEST_NODE_URL = "https://ethereum-sepolia-rpc.publicnode.com"

bsc_client = AioTxBSCClient(BSC_TEST_NODE_URL)
eth_client = AioTxETHClient(ETH_TEST_NODE_URL)


@bsc_client.monitor.on_block
async def handle_block(block):
    print("bsc_client: block", block)


@bsc_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("bsc_client: transaction", transaction)


@eth_client.monitor.on_block
async def handle_block(block):
    print("eth_client: block", block)


@eth_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("eth_client: transaction", transaction)


async def main():
    monitoring_task1 = asyncio.create_task(bsc_client.start_monitoring(584))
    monitoring_task2 = asyncio.create_task(eth_client.start_monitoring(900))
    await asyncio.gather(monitoring_task1, monitoring_task2)

    try:
        while True:
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        bsc_client.stop_monitoring()
        eth_client.stop_monitoring()


if __name__ == "__main__":
    asyncio.run(main())



from ._address import Address
from ._currency import TonCurrencyEnum, from_nano, to_nano
from ._exceptions import InvalidAddressError
from ._utils import (
    b64str_to_bytes,
    b64str_to_hex,
    bytes_to_b64str,
    compare_bytes,
    concat_bytes,
    crc16,
    crc32c,
    move_to_end,
    read_n_bytes_uint_from_array,
    sign_message,
    tree_walk,
)

__all__ = [
    "Address",
    "InvalidAddressError",
    "concat_bytes",
    "move_to_end",
    "tree_walk",
    "crc32c",
    "crc16",
    "read_n_bytes_uint_from_array",
    "compare_bytes",
    "sign_message",
    "b64str_to_bytes",
    "b64str_to_hex",
    "bytes_to_b64str",
    "to_nano",
    "from_nano",
    "TonCurrencyEnum",
]



import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxBSCClient


@vcr_c.use_cassette("tests/fixtures/cassettes/tron/test_async_monitoring.yaml")
async def test_async_monitoring(tron_client: AioTxBSCClient):
    blocks = []
    transactions = []

    @tron_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @tron_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await tron_client.start_monitoring(44739341)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        tron_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 44739341 in blocks
    assert 44739342 in blocks
    assert 44739343 in blocks

    # TRX transfer from block 44739341
    assert "0xb05b7aea5a92a03482af7b764a3c75c0c54de169abdff79d9cb4c4b545fa97a1" in [
        tx["hash"] for tx in transactions
    ]

    # TRC20 transaction from block 44739342
    assert "0x18ea751dac569c9143d679cd8f4c53680c94e336a1b54008b940212a54fbc491" in [
        tx["hash"] for tx in transactions
    ]
    assert "0xec4841f9675a0785c52cb8d7d2c3ca0aeedd4df4590ef4a94b4ca21b48ef11ae" in [
        tx["hash"] for tx in transactions
    ]

    # TRC20 transaction from block 44739343
    assert "0x48a89e3e068379761fdf2d0e2088ffe880803ac8de9e70632fc620b1450045ca" in [
        tx["hash"] for tx in transactions
    ]

    for tx in transactions:
        assert "aiotx_decoded_input" in tx.keys()



import asyncio

from aiotx.clients import AioTxETHClient

eth_client = AioTxETHClient(node_url="https://ethereum-sepolia-rpc.publicnode.com")

private_key = "231d9a16df21402be2b2b50343c489129f1d42dcbaa18b553e6c99057e699df6"
USDC_contract = "0x13fA158A117b93C27c55b8216806294a0aE88b6D"
USDT_contract = "0x419Fe9f14Ff3aA22e46ff1d03a73EdF3b70A62ED"
to_address = "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a"

eth_in_wei = eth_client.to_wei(0.00001, "ether")
tokens_in_mwei = eth_client.to_wei(1, "mwei")


async def main():
    nonce = await eth_client.get_transactions_count(to_address)
    tx_id = await eth_client.send(private_key, to_address, eth_in_wei)
    print(tx_id)
    tx_id = await eth_client.send_token(
        private_key, to_address, USDT_contract, tokens_in_mwei, nonce=nonce + 1
    )
    print(tx_id)
    tx_id = await eth_client.send_token(
        private_key, to_address, USDT_contract, tokens_in_mwei, nonce=nonce + 2
    )
    print(tx_id)


asyncio.run(main())



from ....boc import Cell
from ... import Contract


class NFTSale(Contract):
    code = "B5EE9C7241020A010001B4000114FF00F4A413F4BCF2C80B01020120020302014804050004F2300202CD0607002FA03859DA89A1F481F481F481F401A861A1F401F481F4006101F7D00E8698180B8D8492F82707D201876A2687D207D207D207D006A18116BA4E10159C71D991B1B2990E382C92F837028916382F970FA01698FC1080289C6C8895D7970FAE99F98FD2018201A642802E78B2801E78B00E78B00FD016664F6AA701363804C9B081B2299823878027003698FE99F9810E000C92F857010C0801F5D41081DCD650029285029185F7970E101E87D007D207D0018384008646582A804E78B28B9D090D0A85AD08A500AFD010AE5B564B8FD80384008646582AC678B2803FD010B65B564B8FD80384008646582A802E78B00FD0109E5B564B8FD80381041082FE61E8A10C00C646582A802E78B117D010A65B509E58F8A40900C8C0029A3110471036454012F004E032363704C0038E4782103B9ACA0015BEF2E1C95312C70559C705B1F2E1CA702082105FCC3D14218010C8CB055006CF1622FA0215CB6A14CB1F14CB3F21CF1601CF16CA0021FA02CA00C98100A0FB00E05F06840FF2F0002ACB3F22CF1658CF16CA0021FA02CA00C98100A0FB00AECABAD1"

    def __init__(self, **kwargs):
        self.code = kwargs.get("code") or self.code
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)

    def create_data_cell(self) -> Cell:
        cell = Cell()
        cell.bits.write_address(self.options["marketplace_address"])
        cell.bits.write_address(self.options["nft_address"])
        cell.bits.write_address(None)  # nft_owner_address
        cell.bits.write_grams(self.options["full_price"])

        fees_cell = Cell()
        fees_cell.bits.write_coins(self.options["marketplace_fee"])
        fees_cell.bits.write_address(self.options["royalty_address"])
        fees_cell.bits.write_coins(self.options["royalty_amount"])
        cell.refs.append(fees_cell)
        return cell

    def create_cancel_body(self, query_id: int = 0) -> Cell:
        cell = Cell()
        cell.bits.write_uint(3, 32)  # cancel OP-code
        cell.bits.write_uint(query_id, 64)
        return cell



import os

import pytest
from conftest import vcr_c

from aiotx.clients import AioTxTRONClient
from aiotx.exceptions import (
    CreateTransactionError,
    InvalidArgumentError,
    RpcConnectionError,
    TransactionNotFound,
)

TRON_TEST_WALLET_PRIVATE_KEY = os.environ.get("TRON_TEST_WALLET_PRIVATE_KEY")
assert (
    TRON_TEST_WALLET_PRIVATE_KEY is not None
), "Please provide TRON_TEST_WALLET_PRIVATE_KEY"
CONTRACT = "TG3XXyExBkPp9nzdajDZsozEu4BkaSJozs"
DESTINATION_ADDRESS = "TYge3Gid6vVaQvnPVRJ6SVwzC64cw2eBkN"


async def test_wallet_generator_convertor(tron_client: AioTxTRONClient):
    wallet = tron_client.generate_address()
    print("wallet", wallet)

    base_58_from_hex = tron_client.hex_address_to_base58(wallet["hex_address"])
    hex_from_base58 = tron_client.base58_to_hex_address(wallet["base58check_address"])

    wallet_from_key = tron_client.get_address_from_private_key(wallet["private_key"])

    assert wallet == wallet_from_key
    assert base_58_from_hex == wallet["base58check_address"]
    assert hex_from_base58 == wallet["hex_address"]


@vcr_c.use_cassette("tron/get_last_block.yaml")
async def test_get_last_block(tron_client: AioTxTRONClient):
    block_id = await tron_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("tron/get_chain_id.yaml")
async def test_get_chain_id(tron_client: AioTxTRONClient):
    chain_id = await tron_client.get_chain_id()
    assert isinstance(chain_id, int)
    assert chain_id == 2494104990


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_balance",
    [
        ("TFJfLNUX2xbxrmV3YKt8mQJ3vNT6U6Pnc5", None, 2078288163),
        ("TTfKxPqHXDuPsXM7i5oR6uSS8gH2W1UFT7", None, 7058424560),
        ("TMBS8e3toASGy3zgnpgv67rEXXzpiQa9VN", None, 5221937700),
        ("TG8rTPRYkdgYBDi794PK5GGudAP5mMWLye", None, 887415410180),
        ("TGc3KVY27X9gUb5EH9kwbgr9VUv1BxXq5L", None, 99983900000),
        ("TAhMH42Mr61MRLkbEpT1vjbXpsKSXBtHRz", None, 0),
        ("TWkcsRj1FnAXA1HzWZEW93hdoQxQ2YXFN", ValueError, 0),
    ],
)
@vcr_c.use_cassette("tron/get_balance.yaml")
async def test_get_balance(
    tron_client: AioTxTRONClient, wallet_address, expected_exception, expected_balance
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_balance(wallet_address)
    else:
        balance = await tron_client.get_balance(wallet_address)
        assert isinstance(balance, int)
        assert balance == expected_balance


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        ("5590fe214bfd566bdf35b19bea05a4392e895fe114102c162576200c3c2b3fba", None),
        ("af267a574245393ee68eac6e9d66f1f109b1e06b825743ff645e1b561d654364", None),
        ("101187c488a251ac092f846f16ec190ae4e7230a029716e2a4fdeb10e720b3f9", None),
        ("8a9247c3db7073298463285fabc9150041b2dc5e19e79a955211e0953e3e6bc9", None),
        ("579d94e7f8131b53aac2634426dca3a6d84950ce44512092d599cae7f331ae09", None),
        ("d695d2917e138edb0407ae77109072e43998c2dd6c68be229c62aa49f0dc43ac", None),
        (
            "0820f27e5a67ee3443d0b4a60a1f6c6625328b074dedc2f17c0ea8c44ea440e6",
            TransactionNotFound,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b2",
            InvalidArgumentError,
        ),
        (
            "0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf701b72",
            InvalidArgumentError,
        ),
    ],
)
@vcr_c.use_cassette("tron/get_transaction.yaml")
async def test_get_transaction(tron_client: AioTxTRONClient, tx_id, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_transaction(tx_id)
    else:
        tx = await tron_client.get_transaction(tx_id)
        assert "aiotx_decoded_input" in tx.keys()


@pytest.mark.parametrize(
    "wallet_address, contract, expected_exception, expected_balance",
    [
        ("TSNEe5Tf4rnc9zPMNXfaTF5fZfHDDH8oyW", CONTRACT, None, 9858604795634258),
        ("TT5ZmVgbxiPyixb5Y99F4i5vcDygwH1mPm", CONTRACT, None, 88405001000000),
        ("TX9ctsGJSx8LniV1JgPuvNkXdTGrDdyZqt", CONTRACT, None, 0),
        ("TQkqTUS8iTPmRHHA3rQf1bLCoe9DqZVhsR", CONTRACT, None, 1500000000000),
        ("TGYycKuwtyJ1zRt196PhaWfPuUPGr1wi2P", CONTRACT, None, 0),
        ("TGYycKuwtyJ1zRt196PhaWfPuUPGr1wi2D", CONTRACT, ValueError, 0),
        ("TAhMH42Mr61MRLkbEpT1vjbXpsKSXBtHRz", CONTRACT, None, 0),
        (
            "0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C",
            "Eeaa7E347ee4557f9a6F71C",
            InvalidArgumentError,
            0,
        ),
    ],
)
@vcr_c.use_cassette("tron/get_token_balance.yaml")
async def test_get_token_balance(
    tron_client: AioTxTRONClient,
    wallet_address,
    contract,
    expected_exception,
    expected_balance,
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_contract_balance(wallet_address, contract)
    else:
        balance = await tron_client.get_contract_balance(wallet_address, contract)
        assert isinstance(balance, int)
        assert expected_balance == balance


@pytest.mark.parametrize(
    "contract, expected_decimals, expected_exception",
    [
        (CONTRACT, 6, None),
        ("TB5NSkyzxkzi3eHW87NwFE6TmtTmnZw61y", 6, None),
        ("TSos1xxjqMrGKBxycVmtgrnFvv9M6FDFUX", 9, RpcConnectionError),
    ],
)
@vcr_c.use_cassette("tron/get_contract_decimals.yaml")
async def test_get_contract_decimals(
    tron_client: AioTxTRONClient, contract, expected_decimals, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_contract_decimals(contract)
    else:
        decimals = await tron_client.get_contract_decimals(contract)
        assert decimals == expected_decimals


@vcr_c.use_cassette("tron/get_block_by_number.yaml")
async def test_get_block_by_number(tron_client: AioTxTRONClient):
    block = await tron_client.get_block_by_number(44739224)
    assert isinstance(block, dict)

    assert (
        block["hash"]
        == "0x0000000002aaaa98ab6a708d8c1ce596055a222059983895733031582fc5563e"
    )

    tx_hashes = [tx["hash"] for tx in block["transactions"]]
    assert (
        "0x98f419efccbcfef53bafac52f85b4083c1092c72b7324156a0ce4882706065e8"
        in tx_hashes
    )
    assert (
        "0x1e580e6a15956bdf39df75728ca658cc868b4a321399c95af81bf15e8dc47a13"
        in tx_hashes
    )


@pytest.mark.parametrize(
    "amount, memo, send_to, private_key, expected_exception, expected_tx_id",
    [
        (
            1,
            "",
            DESTINATION_ADDRESS,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "53a4f1ef3614b49c530708c109556246bc87faab6d34a428e77c0419ea940041",
        ),
        (1, 5, DESTINATION_ADDRESS, TRON_TEST_WALLET_PRIVATE_KEY, TypeError, ""),
        (
            5,
            "orp3uhg$£(TUP42t38P(HU$h3r5g8))",
            DESTINATION_ADDRESS,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "06227edc693ed47962d1789fecf6bab84fe9fe8d9961a35daf79127c57256ecb",
        ),
        (
            99999999999,
            "test_memo",
            DESTINATION_ADDRESS,
            TRON_TEST_WALLET_PRIVATE_KEY,
            CreateTransactionError,
            "",
        ),
        (1, "test_memo", "d", TRON_TEST_WALLET_PRIVATE_KEY, CreateTransactionError, ""),
        (
            2,
            "test_memo",
            "Yge3Gid6vVaQvnPVRJ6SVwzC64cw2eBkN",
            TRON_TEST_WALLET_PRIVATE_KEY,
            CreateTransactionError,
            "",
        ),
        (1, "test_memo", DESTINATION_ADDRESS, "d", ValueError, ""),
    ],
)
@vcr_c.use_cassette("tron/send_trx.yaml")
async def test_send_trx(
    tron_client: AioTxTRONClient,
    amount,
    memo,
    send_to,
    private_key,
    expected_exception,
    expected_tx_id,
):
    sun_amount = tron_client.to_sun(amount)
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.send(private_key, send_to, sun_amount, memo)
    else:
        tx_id = await tron_client.send(private_key, send_to, sun_amount, memo)
        assert tx_id == expected_tx_id


@pytest.mark.parametrize(
    "amount, memo, send_to, contract, private_key, expected_exception, expected_tx_id",
    [
        (
            1,
            "",
            DESTINATION_ADDRESS,
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "99dce8bfadc68a0388d7ff28702648402d4fe3dd50e916665368c0a0b6a28273",
        ),
        (
            1,
            5,
            DESTINATION_ADDRESS,
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            TypeError,
            "",
        ),
        (
            5,
            "orp3uhg$£(TUP42t38P(HU$h3r5g8))",
            DESTINATION_ADDRESS,
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "f610753e84a5fe9b8fe72e1ac3175d5584887844e87f14cb08d532d0c66ff1ed",
        ),
        (
            99999999999,
            "test_memo",
            DESTINATION_ADDRESS,
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "31dc98c67bdf5d21bda64f4c11c9fc3c629d3ea0dc437314615c852a27f09b8e",
        ),
        (1, "test_memo", "d", CONTRACT, TRON_TEST_WALLET_PRIVATE_KEY, TypeError, ""),
        (
            2,
            "test_memo",
            "Yge3Gid6vVaQvnPVRJ6SVwzC64cw2eBkN",
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            TypeError,
            "",
        ),
        (1, "test_memo", DESTINATION_ADDRESS, CONTRACT, "d", ValueError, ""),
        (
            99999999999,
            "",
            DESTINATION_ADDRESS,
            CONTRACT,
            TRON_TEST_WALLET_PRIVATE_KEY,
            None,
            "ff0c8255eda1cf21d3e3f150570e1798a7c0a320490b00623601ace17a769bb2",
        ),
        (
            1,
            "",
            DESTINATION_ADDRESS,
            "",
            TRON_TEST_WALLET_PRIVATE_KEY,
            CreateTransactionError,
            "53a4f1ef3614b49c530708c109556246bc87faab6d34a428e77c0419ea940041",
        ),
        (
            1,
            "",
            DESTINATION_ADDRESS,
            "d",
            TRON_TEST_WALLET_PRIVATE_KEY,
            CreateTransactionError,
            "53a4f1ef3614b49c530708c109556246bc87faab6d34a428e77c0419ea940041",
        ),
    ],
)
@vcr_c.use_cassette("tron/test_send_trc20_token.yaml")
async def test_send_trc20_token(
    tron_client: AioTxTRONClient,
    amount,
    memo,
    send_to,
    contract,
    private_key,
    expected_exception,
    expected_tx_id,
):
    # Because token is having 6 decimals too as TRX but check you token!
    sun_amount = tron_client.to_sun(amount)
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.send_token(
                private_key, send_to, contract, sun_amount, memo
            )
    else:
        tx_id = await tron_client.send_token(
            private_key, send_to, contract, sun_amount, memo
        )
        assert tx_id == expected_tx_id



from ._builder import Builder, begin_cell
from ._cell import Cell, deserialize_cell_data, parse_boc_header
from ._dict_builder import DictBuilder, begin_dict

__all__ = [
    "Cell",
    "Builder",
    "begin_cell",
    "DictBuilder",
    "begin_dict",
    "deserialize_cell_data",
    "parse_boc_header",
]



import asyncio

from aiotx.clients import AioTxLTCClient

LTC_TEST_NODE_URL = "TESTNET NODE URL"
ltc_client = AioTxLTCClient(LTC_TEST_NODE_URL, testnet=True)


private_key = "18553b3eb4c5f905d7b023e1736b55ffcbf47657d2cb27f398526ddc9d1764e7"
address = "tltc1qekukv0c9frj3zz8jyag863p8gvv7m3gy2np88d"
to_address = "tltc1qswa8y94sd0njs0atf7h2rmr638nuksw0p2m03v"


async def main():
    balance = await ltc_client.get_balance(
        "tltc1qekukv0c9frj3zz8jyag863p8gvv7m3gy2np88d"
    )
    print("balance is:", balance)
    amount_in_satoshi = ltc_client.to_satoshi(0.0005)
    # Without any additional params
    tx_id = await ltc_client.send(private_key, to_address, amount_in_satoshi)
    print(tx_id)
    # Deduct fee from client
    tx_id = await ltc_client.send(
        private_key, to_address, amount_in_satoshi, deduct_fee=True
    )
    print(tx_id)
    tx_id = await ltc_client.send_bulk(
        private_key,
        {
            to_address: amount_in_satoshi,
            "tltc1qswslzcdulvlk62gdrg8wa0sw36f938h2cvtaf7": amount_in_satoshi,
        },
    )
    print(tx_id)
    balance = await ltc_client.get_balance(
        "tltc1qekukv0c9frj3zz8jyag863p8gvv7m3gy2np88d"
    )
    print("balance is:", balance)


asyncio.run(main())



from enum import Enum
from typing import List, Optional, Tuple

from ...crypto import mnemonic_is_valid, mnemonic_new, mnemonic_to_wallet_key
from ...crypto.exceptions import InvalidMnemonicsError
from ._highload_wallet_contract import HighloadWalletV2Contract
from ._multisig_wallet_contract import (
    MultiSigOrder,
    MultiSigOrderBuilder,
    MultiSigWallet,
)
from ._wallet_contract import SendModeEnum, WalletContract
from ._wallet_contract_v2 import WalletV2ContractR1, WalletV2ContractR2
from ._wallet_contract_v3 import WalletV3ContractR1, WalletV3ContractR2
from ._wallet_contract_v4 import WalletV4ContractR1, WalletV4ContractR2


class WalletVersionEnum(str, Enum):
    v2r1 = "v2r1"
    v2r2 = "v2r2"
    v3r1 = "v3r1"
    v3r2 = "v3r2"
    v4r1 = "v4r1"
    v4r2 = "v4r2"
    hv2 = "hv2"


class Wallets:
    default_version = WalletVersionEnum.v3r2
    ALL = {
        WalletVersionEnum.v2r1: WalletV2ContractR1,
        WalletVersionEnum.v2r2: WalletV2ContractR2,
        WalletVersionEnum.v3r1: WalletV3ContractR1,
        WalletVersionEnum.v3r2: WalletV3ContractR2,
        WalletVersionEnum.v4r1: WalletV4ContractR1,
        WalletVersionEnum.v4r2: WalletV4ContractR2,
        WalletVersionEnum.hv2: HighloadWalletV2Contract,
    }

    @classmethod
    def create(
        cls,
        version: WalletVersionEnum,
        workchain: int,
        password: Optional[str] = None,
        **kwargs,
    ) -> Tuple[List[str], bytes, bytes, WalletContract]:
        """
        :rtype: (List[str](mnemonics), bytes(public_key), bytes(private_key), WalletContract(wallet))
        """
        mnemonics = mnemonic_new(password=password)
        pub_k, priv_k = mnemonic_to_wallet_key(mnemonics)
        wallet = cls.ALL[version](
            public_key=pub_k, private_key=priv_k, wc=workchain, **kwargs
        )

        return mnemonics, pub_k, priv_k, wallet

    @classmethod
    def from_mnemonics(
        cls,
        mnemonics: List[str],
        version: WalletVersionEnum = default_version,
        workchain: int = 0,
        **kwargs,
    ) -> Tuple[List[str], bytes, bytes, WalletContract]:
        """
        :rtype: (List[str](mnemonics), bytes(public_key), bytes(private_key), WalletContract(wallet))
        """
        if not mnemonic_is_valid(mnemonics):
            raise InvalidMnemonicsError()

        pub_k, priv_k = mnemonic_to_wallet_key(mnemonics)
        wallet = cls.ALL[version](
            public_key=pub_k, private_key=priv_k, wc=workchain, **kwargs
        )

        return mnemonics, pub_k, priv_k, wallet

    @classmethod
    def to_addr_pk(
        cls,
        mnemonics: WalletContract,
        version: WalletVersionEnum = default_version,
        workchain: int = 0,
        **kwargs,
    ) -> Tuple[bytes, bytes]:
        """
        :rtype: (bytes(addr), bytes(pk)
        """
        _mnemonics, _pub_k, priv_k, wallet = cls.from_mnemonics(
            mnemonics, version, workchain, **kwargs
        )

        return wallet.address.to_buffer(), priv_k[:32]


__all__ = [
    "WalletV2ContractR1",
    "WalletV2ContractR2",
    "WalletV3ContractR1",
    "WalletV3ContractR2",
    "WalletV4ContractR1",
    "WalletV4ContractR2",
    "WalletContract",
    "SendModeEnum",
    "WalletVersionEnum",
    "Wallets",
    "MultiSigWallet",
    "MultiSigOrder",
    "MultiSigOrderBuilder",
]



import hashlib
import hmac
from typing import List, Optional, Tuple

from nacl.bindings import crypto_sign_seed_keypair

from ._settings import PBKDF_ITERATIONS
from ._utils import get_secure_random_number, is_basic_seed
from .bip39 import english


def mnemonic_is_valid(mnemo_words: List[str]) -> bool:
    return len(mnemo_words) == 24 and is_basic_seed(mnemonic_to_entropy(mnemo_words))


def mnemonic_to_entropy(mnemo_words: List[str], password: Optional[str] = None):
    # TODO: implement password
    sign = hmac.new(
        (" ".join(mnemo_words)).encode("utf-8"), bytes(0), hashlib.sha512
    ).digest()
    return sign


def mnemonic_to_seed(mnemo_words: List[str], seed: str, password: Optional[str] = None):
    entropy = mnemonic_to_entropy(mnemo_words, password)
    return hashlib.pbkdf2_hmac("sha512", entropy, seed, PBKDF_ITERATIONS)


def mnemonic_to_private_key(
    mnemo_words: List[str], password: Optional[str] = None
) -> Tuple[bytes, bytes]:
    """
    :rtype: (bytes(public_key), bytes(secret_key))
    """
    seed = mnemonic_to_seed(mnemo_words, "TON default seed".encode("utf-8"), password)
    return crypto_sign_seed_keypair(seed[:32])


def mnemonic_to_wallet_key(
    mnemo_words: List[str], password: Optional[str] = None
) -> Tuple[bytes, bytes]:
    """
    :rtype: (bytes(public_key), bytes(secret_key))
    """
    _, priv_k = mnemonic_to_private_key(mnemo_words, password)
    return crypto_sign_seed_keypair(priv_k[:32])


def mnemonic_new(words_count: int = 24, password: Optional[str] = None) -> List[str]:
    while True:
        mnemo_arr = []

        for _ in range(words_count):
            idx = get_secure_random_number(0, len(english))
            mnemo_arr.append(english[idx])

        if not is_basic_seed(mnemonic_to_entropy(mnemo_arr)):
            continue

        break

    return mnemo_arr



import json

import pkg_resources

from ._evm_base_client import AioTxEVMClient
from ._ton_base_client import AioTxTONClient
from ._tron_base_client import AioTxTRONClient
from ._utxo_base_client import AioTxUTXOClient

__all__ = [
    "AioTxTONClient",
    "AioTxBSCClient",
    "AioTxETHClient",
    "AioTxPolygonClient",
    "AioTxBTCClient",
    "AioTxLTCClient",
    "AioTxTRONClient",
]


class AioTxBSCClient(AioTxEVMClient):
    def __init__(self, node_url: str, headers: dict = {}):
        super().__init__(node_url, headers)
        bep20_abi_json = pkg_resources.resource_string("aiotx.utils", "bep20_abi.json")
        self._bep20_abi = json.loads(bep20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._bep20_abi]


class AioTxETHClient(AioTxEVMClient):
    def __init__(self, node_url: str, headers: dict = {}):
        super().__init__(node_url, headers)
        erc20_abi_json = pkg_resources.resource_string("aiotx.utils", "erc20_abi.json")
        self._erc20_abi = json.loads(erc20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._erc20_abi]


class AioTxPolygonClient(AioTxEVMClient):
    def __init__(self, node_url: str, headers: dict = {}):
        super().__init__(node_url, headers)
        erc20_abi_json = pkg_resources.resource_string("aiotx.utils", "erc20_abi.json")
        self._erc20_abi = json.loads(erc20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._erc20_abi]


class AioTxBTCClient(AioTxUTXOClient):
    def __init__(
        self,
        node_url,
        headers: dict = {},
        testnet=False,
        node_username: str = "",
        node_password: str = "",
        db_url="sqlite+aiosqlite:///aiotx_utxo.sqlite",
    ):
        network_name = "testnet" if testnet else "bitcoin"
        super().__init__(
            node_url,
            headers,
            testnet,
            node_username,
            node_password,
            network_name,
            db_url,
        )


class AioTxLTCClient(AioTxUTXOClient):
    def __init__(
        self,
        node_url,
        headers: dict = {},
        testnet=False,
        node_username: str = "",
        node_password: str = "",
        db_url="sqlite+aiosqlite:///aiotx_utxo.sqlite",
    ):
        network_name = "litecoin_testnet" if testnet else "litecoin"
        super().__init__(
            node_url,
            headers,
            testnet,
            node_username,
            node_password,
            network_name,
            db_url,
        )



import decimal
from enum import Enum
from typing import Any, Union


class TonCurrencyEnum(str, Enum):
    nanoton = "nanoton"
    ton = "ton"


units = {
    TonCurrencyEnum.nanoton: decimal.Decimal("1"),
    TonCurrencyEnum.ton: decimal.Decimal("1000000000"),
}

integer_types = (int,)
string_types = (bytes, str, bytearray)

MIN_VAL = 0
MAX_VAL = 2**256 - 1


def is_integer(value: Any) -> bool:
    return isinstance(value, integer_types) and not isinstance(value, bool)


def is_string(value: Any) -> bool:
    return isinstance(value, string_types)


def to_nano(number: Union[int, float, str, decimal.Decimal], unit: str) -> int:
    """from coins to nanocoins

    Args:
        number (Union[int, float, str, decimal.Decimal])
        unit (str): unit of the number

    Returns:
        int: nanoton
    """
    if unit.lower() not in units:
        raise ValueError(
            "Unknown unit.  Must be one of {0}".format("/".join(units.keys()))
        )

    if is_integer(number) or is_string(number):
        d_number = decimal.Decimal(value=number)
    elif isinstance(number, float):
        d_number = decimal.Decimal(value=str(number))
    elif isinstance(number, decimal.Decimal):
        d_number = number
    else:
        raise TypeError("Unsupported type.  Must be one of integer, float, or string")

    s_number = str(number)
    unit_value = units[unit.lower()]

    if d_number == decimal.Decimal(0):
        return 0

    if d_number < 1 and "." in s_number:
        with decimal.localcontext() as ctx:
            multiplier = len(s_number) - s_number.index(".") - 1
            ctx.prec = multiplier
            d_number = decimal.Decimal(value=number, context=ctx) * 10**multiplier
        unit_value /= 10**multiplier

    with decimal.localcontext() as ctx:
        ctx.prec = 999
        result_value = decimal.Decimal(value=d_number, context=ctx) * unit_value

    if result_value < MIN_VAL or result_value > MAX_VAL:
        raise ValueError("Resulting nanoton value must be between 1 and 2**256 - 1")

    return int(result_value)


def from_nano(number: int, unit: str) -> Union[int, decimal.Decimal]:
    """from nanocoins to coins

    Args:
        number (int)
        unit (str): required unit

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        Union[int, decimal.Decimal]: _description_
    """
    if unit.lower() not in units:
        raise ValueError(
            "Unknown unit.  Must be one of {0}".format("/".join(units.keys()))
        )

    if number == 0:
        return 0

    if number < MIN_VAL or number > MAX_VAL:
        raise ValueError("value must be between 1 and 2**256 - 1")

    unit_value = units[unit.lower()]

    with decimal.localcontext() as ctx:
        ctx.prec = 999
        d_number = decimal.Decimal(value=number, context=ctx)
        result_value = d_number / unit_value

    return result_value
